"""
Import the functions from the implementations into the public API
"""

from scores.continuous.flip_flop_impl import (
    flip_flop_index,
    flip_flop_index_proportion_exceeding,
)
from scores.continuous.murphy_impl import murphy_score, murphy_thetas
from scores.continuous.quantile_loss_impl import quantile_score
from scores.continuous.standard_impl import (
    additive_bias,
    correlation,
    mae,
    mean_error,
    mse,
    multiplicative_bias,
    rmse,
)
from scores.processing.isoreg_impl import isotonic_fit

__all__ = [
    "flip_flop_index",
    "flip_flop_index_proportion_exceeding",
    "murphy_score",
    "murphy_thetas",
    "quantile_score",
    "correlation",
    "mae",
    "mse",
    "rmse",
    "additive_bias",
    "mean_error",
    "multiplicative_bias",
    "isotonic_fit",
]



"""
This foundation class provides an underpinning data structure which can be used for 
contingency tables of various kinds, also known as a confusion matrix.

It allows the careful setting out of when and where forecasts closely match
observations.

The binary contingency table captures true positives (hits), true negatives (correct negatives), 
false positives (false alarms) and false negatives (misses). 

The process of deriving a contingency table relies on the forecast data, the observation
data, and a matching or event operator. The event operator will produce the category from the
forecast and the observed data. Specific use cases may well be supported by a case-specific
matching operator.

Scores supports complex, weighted, multi-dimensional data, including in contingency tables.

Users can supply their own event operators to the top-level module functions.
"""

import operator
from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import xarray as xr

import scores.utils
from scores.typing import FlexibleArrayType, FlexibleDimensionTypes

DEFAULT_PRECISION = 8


class BasicContingencyManager:
    """
    A BasicContingencyManager is produced when a BinaryContingencyManager is transformed.

    A basic contingency table is built only from the event counts, losing the connection
    to the actual event tables in their full dimensionality.

    The event count data is much smaller than the full event tables, particularly when
    considering very large data sets like NWP data, which could be terabytes to petabytes
    in size.
    """

    def __init__(self, counts: dict):
        """
        Compute any arrays required and store the resulting counts.
        """
        for key, arr in counts.items():
            counts[key] = arr.compute()

        self.counts = counts
        self._make_xr_table()

    def _make_xr_table(
        self,
    ):
        """
        From a dictionary of the skill score elements, produce an xarray
        structure which corresponds.
        """
        ctable_list = []
        entry_name = []
        for i, j in self.counts.items():
            entry_name.append(i)
            ctable_list.append(j)

        xr_table = xr.concat(ctable_list, dim="contingency")
        xr_table["contingency"] = entry_name

        self.xr_table = xr_table

    def __str__(self):
        heading = "Contingency Manager (xarray view of table):"
        table = self.xr_table
        tablerepr = repr(table)
        final = "\n".join([heading, tablerepr])
        return final

    def get_counts(self):
        """
        Return the contingency table counts (tp, fp, tn, fn)
        """
        return self.counts

    def get_table(self):
        """
        Return the contingency table as an xarray object
        """
        return self.xr_table

    def accuracy(self):
        """
        The proportion of forecasts which are true

        https://www.cawcr.gov.au/projects/verification/#ACC
        """
        count_dictionary = self.counts
        correct_count = count_dictionary["tp_count"] + count_dictionary["tn_count"]
        ratio = correct_count / count_dictionary["total_count"]
        return ratio

    def frequency_bias(self):
        """
        How did the forecast frequency of "yes" events compare to the observed frequency of "yes" events?

        https://www.cawcr.gov.au/projects/verification/#BIAS
        """
        # Note - bias_score calls this method
        cd = self.counts
        freq_bias = (cd["tp_count"] + cd["fp_count"]) / (cd["tp_count"] + cd["fn_count"])

        return freq_bias

    def bias_score(self):
        """
        How did the forecast frequency of "yes" events compare to the observed frequency of "yes" events?

        https://www.cawcr.gov.au/projects/verification/#BIAS
        """
        return self.frequency_bias()

    def hit_rate(self):
        """
        What proportion of the observed events where correctly forecast?
        Identical to probability_of_detection
        Range: 0 to 1.  Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#POD
        """
        return self.probability_of_detection()

    def probability_of_detection(self):
        """
        What proportion of the observed events where correctly forecast?
        Identical to hit_rate
        Range: 0 to 1.  Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#POD
        """
        # Note - hit_rate and sensitiviy call this function
        cd = self.counts
        pod = cd["tp_count"] / (cd["tp_count"] + cd["fn_count"])

        return pod

    def true_positive_rate(self):
        """
        What proportion of the observed events where correctly forecast?
        Identical to probability_of_detection
        Range: 0 to 1.  Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#POD
        """
        return self.probability_of_detection()

    def false_alarm_ratio(self):
        """
        What fraction of the predicted "yes" events actually did not occur (i.e.,
        were false alarms)?
        Range: 0 to 1. Perfect score: 0.

        https://www.cawcr.gov.au/projects/verification/#FAR
        """
        cd = self.counts
        far = cd["fp_count"] / (cd["tp_count"] + cd["fp_count"])

        return far

    def false_alarm_rate(self):
        """
        What fraction of the non-events were incorrectly predicted?
        Identical to probability_of_false_detection
        Range: 0 to 1.  Perfect score: 0.

        https://www.cawcr.gov.au/projects/verification/#POFD
        """
        # Note - probability of false detection calls this function
        cd = self.counts
        far = cd["fp_count"] / (cd["tn_count"] + cd["fp_count"])

        return far

    def probability_of_false_detection(self):
        """
        What fraction of the non-events were incorrectly predicted?
        Identical to false_alarm_rate
        Range: 0 to 1.  Perfect score: 0.

        https://www.cawcr.gov.au/projects/verification/#POFD
        """

        return self.false_alarm_rate()

    def success_ratio(self):
        """
        What proportion of the forecast events actually eventuated?
        Range: 0 to 1.  Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#SR
        """
        cd = self.counts
        sr = cd["tp_count"] / (cd["tp_count"] + cd["fp_count"])

        return sr

    def threat_score(self):
        """
        How well did the forecast "yes" events correspond to the observed "yes" events?
        Identical to critical_success_index
        Range: 0 to 1, 0 indicates no skill. Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#CSI
        """
        # Note - critical success index just calls this method

        cd = self.counts
        ts = cd["tp_count"] / (cd["tp_count"] + cd["fp_count"] + cd["tn_count"])
        return ts

    def critical_success_index(self):
        """
        Often known as CSI.

        How well did the forecast "yes" events correspond to the observed "yes" events?
        Identical to threat_score
        Range: 0 to 1, 0 indicates no skill. Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#CSI
        """
        return self.threat_score()

    def peirce_skill_score(self):
        """
        Hanssen and Kuipers discriminant (true skill statistic, Peirce's skill score)
        How well did the forecast separate the "yes" events from the "no" events?
        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#HK
        """
        cd = self.counts
        component_a = cd["tp_count"] / (cd["tp_count"] + cd["fn_count"])
        component_b = cd["fn_count"] / (cd["fn_count"] + cd["tn_count"])
        skill_score = component_a - component_b
        return skill_score

    def true_skill_statistic(self):
        """
        Identical to Peirce's skill score and to Hanssen and Kuipers discriminant
        How well did the forecast separate the "yes" events from the "no" events?
        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#HK
        """
        return self.peirce_skill_score()

    def hanssen_and_kuipers_discriminant(self):
        """
        Identical to Peirce's skill score and to true skill statistic
        How well did the forecast separate the "yes" events from the "no" events?
        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.

        https://www.cawcr.gov.au/projects/verification/#HK
        """
        return self.peirce_skill_score()

    def sensitivity(self):
        """
        https://en.wikipedia.org/wiki/Sensitivity_and_specificity
        """
        return self.probability_of_detection()

    def specificity(self):
        """
        https://en.wikipedia.org/wiki/Sensitivity_and_specificity
        """
        cd = self.counts
        s = cd["tn_count"] / (cd["tn_count"] + cd["fp_count"])
        return s

    def recall(self):
        """
        Identical to probability of detection.

        https://en.wikipedia.org/wiki/Precision_and_recall
        """
        return self.probability_of_detection()

    def precision(self):
        """
        Identical to the Success Ratio.

        https://en.wikipedia.org/wiki/Precision_and_recall
        """
        return self.success_ratio()

    def f1_score(self):
        """
        Calculates the F1 score.
        https://en.wikipedia.org/wiki/F-score
        """
        cd = self.counts
        f1 = 2 * cd["tp_count"] / (2 * cd["tp_count"] + cd["fp_count"] + cd["fn_count"])
        return f1

    def equitable_threat_score(self):
        """
        Calculates the Equitable threat score (also known as the Gilbert skill score).

        How well did the forecast "yes" events correspond to the observed "yes"
        events (accounting for hits due to chance)?

        Range: -1/3 to 1, 0 indicates no skill. Perfect score: 1.

        Hogan, R.J., Ferro, C.A., Jolliffe, I.T. and Stephenson, D.B., 2010.
        Equitability revisited: Why the “equitable threat score” is not equitable.
        Weather and Forecasting, 25(2), pp.710-726.
        """
        cd = self.counts
        hits_random = (cd["tp_count"] + cd["fn_count"]) * (cd["tp_count"] + cd["fp_count"]) / cd["total_count"]
        ets = (cd["tp_count"] - hits_random) / (cd["tp_count"] + cd["fn_count"] + cd["fp_count"] - hits_random)

        return ets

    def gilberts_skill_score(self):
        """
        Calculates the Gilbert skill score (also known as the Equitable threat score).

        How well did the forecast "yes" events correspond to the observed "yes"
        events (accounting for hits due to chance)?

        Range: -1/3 to 1, 0 indicates no skill. Perfect score: 1.

        Hogan, R.J., Ferro, C.A., Jolliffe, I.T. and Stephenson, D.B., 2010.
        Equitability revisited: Why the “equitable threat score” is not equitable.
        Weather and Forecasting, 25(2), pp.710-726.
        """
        return self.equitable_threat_score()

    def heidke_skill_score(self):
        """
        Calculates the Heidke skill score (also known as Cohen's kappa).

        What was the accuracy of the forecast relative to that of random chance?

        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.

        https://en.wikipedia.org/wiki/Cohen%27s_kappa
        """
        cd = self.counts
        exp_correct = (1 / cd["total_count"]) * (
            (cd["tp_count"] + cd["fn_count"]) * (cd["tp_count"] + cd["fp_count"])
            + ((cd["tn_count"] + cd["fn_count"]) * (cd["tn_count"] + cd["fp_count"]))
        )
        hss = ((cd["tp_count"] + cd["tn_count"]) - exp_correct) / (cd["total_count"] - exp_correct)
        return hss

    def cohens_kappa(self):
        """
        Calculates the Cohen's kappa (also known as the Heidke skill score).

        What was the accuracy of the forecast relative to that of random chance?

        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.

        https://en.wikipedia.org/wiki/Cohen%27s_kappa
        """
        return self.heidke_skill_score()

    def odds_ratio(self):
        """
        Calculates the odds ratio

        What is the ratio of the odds of a "yes" forecast being correct, to the odds of
        a "yes" forecast being wrong?

        Odds ratio - Range: 0 to ∞, 1 indicates no skill. Perfect score: ∞.

        Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill.
        Weather and Forecasting, 15(2), pp.221-232.
        """
        odds_r = (self.probability_of_detection() / (1 - self.probability_of_detection())) / (
            self.probability_of_false_detection() / (1 - self.probability_of_false_detection())
        )
        return odds_r

    def odds_ratio_skill_score(self):
        """
        Calculates the odds ratio skill score (also known as Yule's Q).

        Note - the term 'skill score' is often used to describe the relative performance
        of one source of predictoins over another - e.g. the relative performance of an
        upgraded model on its predecessor, or the relative performance to a benchmark such
        as climatology. The odds ratio skill score is not that kind of skill score.
        What was the improvement of the forecast over random chance?
        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
        Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill.
        Weather and Forecasting, 15(2), pp.221-232.
        """
        cd = self.counts
        orss = (cd["tp_count"] * cd["tn_count"] - cd["fn_count"] * cd["fp_count"]) / (
            cd["tp_count"] * cd["tn_count"] + cd["fn_count"] * cd["fp_count"]
        )
        return orss

    def yules_q(self):
        """
        Calculates the Yule's Q (also known as the odds ratio skill score).
        What was the improvement of the forecast over random chance?
        Range: -1 to 1, 0 indicates no skill. Perfect score: 1.
        Stephenson, D.B., 2000. Use of the “odds ratio” for diagnosing forecast skill.
        Weather and Forecasting, 15(2), pp.221-232.
        """
        return self.odds_ratio_skill_score()


class BinaryContingencyManager(BasicContingencyManager):
    """
    At each location, the value will either be:
     - A true positive (hit)
     - A false positive (false alarm)
     - A true negative (correct negative)
     - A false negative (miss)

    It will be common to want to operate on masks of these values,
    such as:
     - Plotting these attributes on a map
     - Calculating the total number of these attributes
     - Calculating various ratios of these attributes, potentially
       masked by geographical area (e.g. accuracy in a region)

    As such, the per-pixel information is useful as well as the overall
    ratios involved.

    BinaryContingencyManager utilises the BasicContingencyManager class to provide
    most functionality.
    """

    def __init__(
        self, forecast_events: FlexibleArrayType, observed_events: FlexibleArrayType
    ):  # pylint: disable=super-init-not-called
        self.forecast_events = forecast_events
        self.observed_events = observed_events

        self.tp = (self.forecast_events == 1) & (self.observed_events == 1)  # true positives
        self.tn = (self.forecast_events == 0) & (self.observed_events == 0)  # true negatives
        self.fp = (self.forecast_events == 1) & (self.observed_events == 0)  # false positives
        self.fn = (self.forecast_events == 0) & (self.observed_events == 1)  # false negatives

        # Bring back NaNs where there is either a forecast or observed event nan
        self.tp = self.tp.where(~np.isnan(forecast_events))
        self.tp = self.tp.where(~np.isnan(observed_events))
        self.tn = self.tn.where(~np.isnan(forecast_events))
        self.tn = self.tn.where(~np.isnan(observed_events))
        self.fp = self.fp.where(~np.isnan(forecast_events))
        self.fp = self.fp.where(~np.isnan(observed_events))
        self.fn = self.fn.where(~np.isnan(forecast_events))
        self.fn = self.fn.where(~np.isnan(observed_events))

        # Variables for count-based metrics
        self.counts = self._get_counts()
        self._make_xr_table()

    def transform(
        self,
        *,
        reduce_dims: Optional[FlexibleDimensionTypes] = None,
        preserve_dims: Optional[FlexibleDimensionTypes] = None,
    ):
        """
        Calculate and compute the contingency table according to the specified dimensions
        """
        cd = self._get_counts(reduce_dims=reduce_dims, preserve_dims=preserve_dims)
        return BasicContingencyManager(cd)

    def _get_counts(
        self,
        *,
        reduce_dims: Optional[FlexibleDimensionTypes] = None,
        preserve_dims: Optional[FlexibleDimensionTypes] = None,
    ):
        """
        Generates the uncomputed count values
        """

        to_reduce = scores.utils.gather_dimensions(
            self.forecast_events.dims,
            self.observed_events.dims,
            reduce_dims=reduce_dims,
            preserve_dims=preserve_dims,
        )

        cd = {
            "tp_count": self.tp.sum(dim=to_reduce),
            "tn_count": self.tn.sum(dim=to_reduce),
            "fp_count": self.fp.sum(dim=to_reduce),
            "fn_count": self.fn.sum(dim=to_reduce),
        }
        total = cd["tp_count"] + cd["tn_count"] + cd["fp_count"] + cd["fn_count"]
        cd["total_count"] = total

        return cd


class EventOperator(ABC):
    """
    Base class for event operators which can be used in deriving contingency
    tables. This will be expanded as additional use cases are incorporated
    beyond the ThresholdEventOperator.
    """

    @abstractmethod
    def make_event_tables(
        self, forecast: FlexibleArrayType, observed: FlexibleArrayType, *, event_threshold=None, op_fn=operator.gt
    ):
        """
        This method should be over-ridden to return forecast and observed event tables
        """
        ...  # pragma: no cover # pylint: disable=unnecessary-ellipsis

    @abstractmethod
    def make_contingency_manager(
        self, forecast: FlexibleArrayType, observed: FlexibleArrayType, *, event_threshold=None, op_fn=operator.gt
    ):
        """
        This method should be over-ridden to return a contingency table.
        """
        ...  # pragma: no cover # pylint: disable=unnecessary-ellipsis


class ThresholdEventOperator(EventOperator):
    """
    Given a forecast and and an observation, consider an event defined by
    particular variables meeting a threshold condition (e.g. rainfall above 1mm).

    This class abstracts that concept for any event definition.
    """

    def __init__(self, *, precision=DEFAULT_PRECISION, default_event_threshold=0.001, default_op_fn=operator.ge):
        self.precision = precision
        self.default_event_threshold = default_event_threshold
        self.default_op_fn = default_op_fn

    def make_event_tables(
        self, forecast: FlexibleArrayType, observed: FlexibleArrayType, *, event_threshold=None, op_fn=None
    ):
        """
        Using this function requires a careful understanding of the structure of the data
        and the use of the operator function. The default operator is a simple greater-than
        operator, so this will work on a simple DataArray. To work on a DataSet, a richer
        understanding is required. It is recommended to work through the Contingency Table
        tutorial to review more complex use cases, including on multivariate gridded model
        data, and on station data structures.
        """

        if not event_threshold:
            event_threshold = self.default_event_threshold

        if not op_fn:
            op_fn = self.default_op_fn

        forecast_events = op_fn(forecast, event_threshold)
        observed_events = op_fn(observed, event_threshold)

        # Bring back NaNs
        forecast_events = forecast_events.where(~np.isnan(forecast))
        observed_events = observed_events.where(~np.isnan(observed))

        return (forecast_events, observed_events)

    def make_contingency_manager(
        self, forecast: FlexibleArrayType, observed: FlexibleArrayType, *, event_threshold=None, op_fn=operator.gt
    ):
        """
        Using this function requires a careful understanding of the structure of the data
        and the use of the operator function. The default operator is a simple greater-than
        operator, so this will work on a simple DataArray. To work on a DataSet, a richer
        understanding is required. It is recommended to work through the Contingency Table
        tutorial to review more complex use cases, including on multivariate gridded model
        data, and on station data structures.
        """

        if not event_threshold:
            event_threshold = self.default_event_threshold

        forecast_events = op_fn(forecast, event_threshold)
        observed_events = op_fn(observed, event_threshold)

        # Bring back NaNs
        forecast_events = forecast_events.where(~np.isnan(forecast))
        observed_events = observed_events.where(~np.isnan(observed))

        table = BinaryContingencyManager(forecast_events, observed_events)
        return table



"""
This module contains standard methods which may be used for continuous scoring
"""

from typing import Optional

import xarray as xr

import scores.functions
import scores.utils
from scores.processing import broadcast_and_match_nan
from scores.typing import FlexibleArrayType, FlexibleDimensionTypes, XarrayLike


def mse(
    fcst: FlexibleArrayType,
    obs: FlexibleArrayType,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
    is_angular: Optional[bool] = False,
) -> XarrayLike:
    """Calculates the mean squared error from forecast and observed data.

    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the metric. At
    most one of reduce_dims and preserve_dims may be specified.
    Specifying both will result in an exception.

    See "Mean squared error" section at https://www.cawcr.gov.au/projects/verification/#MSE for more information

    Args:
        fcst (Union[xr.Dataset, xr.DataArray, pd.Dataframe, pd.Series]):
            Forecast or predicted variables in xarray or pandas.
        obs (Union[xr.Dataset, xr.DataArray, pd.Dataframe, pd.Series]):
            Observed variables in xarray or pandas.
        reduce_dims (Union[str, Iterable[str]): Optionally specify which
            dimensions to reduce when calculating MSE. All other dimensions
            will be preserved.
        preserve_dims (Union[str, Iterable[str]): Optionally specify which
            dimensions to preserve when calculating MSE. All other dimensions
            will be reduced. As a special case, 'all' will allow all dimensions
            to be preserved. In this case, the result will be in the same
            shape/dimensionality as the forecast, and the errors will be
            the squared error at each point (i.e. single-value comparison
            against observed), and the forecast and observed dimensions
            must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.


    Returns:
        Union[xr.Dataset, xr.DataArray, pd.Dataframe, pd.Series]: An object containing
            a single floating point number representing the mean absolute
            error for the supplied data. All dimensions will be reduced.
            Otherwise: Returns an object representing the mean squared error,
            reduced along the relevant dimensions and weighted appropriately.
    """
    if is_angular:
        error = scores.functions.angular_difference(fcst, obs)  # type: ignore
    else:
        error = fcst - obs  # type: ignore
    squared = error * error
    squared = scores.functions.apply_weights(squared, weights=weights)  # type: ignore

    if preserve_dims or reduce_dims:
        reduce_dims = scores.utils.gather_dimensions(
            fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
        )

    if reduce_dims is not None:
        _mse = squared.mean(dim=reduce_dims)  # type: ignore
    else:
        _mse = squared.mean()

    return _mse


def rmse(
    fcst: FlexibleArrayType,
    obs: FlexibleArrayType,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
    is_angular: bool = False,
) -> FlexibleArrayType:
    """Calculate the Root Mean Squared Error from xarray or pandas objects.

    A detailed explanation is on [Wikipedia](https://en.wikipedia.org/wiki/Root-mean-square_deviation)


    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the metric.
    At most one of `reduce_dims` and `preserve_dims` may be specified.
    Specifying both will result in an exception.


    Args:
        fcst: Forecast
            or predicted variables in xarray or pandas.
        obs: Observed
            variables in xarray or pandas.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating RMSE. All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve
            when calculating RMSE. All other dimensions will be reduced.
            As a special case, 'all' will allow all dimensions to be
            preserved. In this case, the result will be in the same
            shape/dimensionality as the forecast, and the errors will be
            the absolute error at each point (i.e. single-value comparison
            against observed), and the forecast and observed dimensions
            must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.

    Returns:
        An object containing
            a single floating point number representing the root mean squared
            error for the supplied data. All dimensions will be reduced.
            Otherwise: Returns an object representing the root mean squared error,
            reduced along the relevant dimensions and weighted appropriately.

    """
    _mse = mse(fcst, obs, reduce_dims=reduce_dims, preserve_dims=preserve_dims, weights=weights, is_angular=is_angular)

    _rmse = pow(_mse, (1 / 2))

    return _rmse  # type: ignore


def mae(
    fcst: FlexibleArrayType,
    obs: FlexibleArrayType,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
    is_angular: bool = False,
) -> FlexibleArrayType:
    """Calculates the mean absolute error from forecast and observed data.

    A detailed explanation is on [Wikipedia](https://en.wikipedia.org/wiki/Mean_absolute_error)

    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the metric.
    At most one of reduce_dims and preserve_dims may be specified.
    Specifying both will result in an exception.

    Args:
        fcst: Forecast or predicted variables in xarray or pandas.
        obs: Observed variables in xarray or pandas.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating MAE. All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating MAE. All other dimensions will be reduced. As a
            special case, 'all' will allow all dimensions to be preserved. In
            this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the absolute error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.

    Returns:
        By default an xarray DataArray containing
        a single floating point number representing the mean absolute error for the
        supplied data. All dimensions will be reduced.

        Alternatively, an xarray structure with dimensions preserved as appropriate
        containing the score along reduced dimensions
    """
    if is_angular:
        error = scores.functions.angular_difference(fcst, obs)  # type: ignore
    else:
        error = fcst - obs  # type: ignore
    ae = abs(error)
    ae = scores.functions.apply_weights(ae, weights=weights)  # type: ignore

    if preserve_dims is not None or reduce_dims is not None:
        reduce_dims = scores.utils.gather_dimensions(
            fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
        )

    if reduce_dims is not None:
        _ae = ae.mean(dim=reduce_dims)
    else:
        _ae = ae.mean()

    # Returns unhinted types if nonstandard types passed in, but this is useful
    return _ae  # type: ignore


def correlation(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
) -> xr.DataArray:
    """
    Calculates the Pearson's correlation coefficient between two xarray DataArrays

    Args:
        fcst: Forecast or predicted variables
        obs: Observed variables.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the Pearson's correlation coefficient.
            All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating the Pearson's correlation coefficient. All other dimensions will
            be reduced. As a special case, 'all' will allow all dimensions to be
            preserved. In this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the absolute error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
    Returns:
        An xarray object with Pearson's correlation coefficient values
    """
    reduce_dims = scores.utils.gather_dimensions(
        fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
    )

    return xr.corr(fcst, obs, reduce_dims)


def mean_error(
    fcst: XarrayLike,
    obs: XarrayLike,
    *,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[XarrayLike] = None,
) -> XarrayLike:
    """
    Calculates the mean error which is also sometimes called the additive bias.

    It is defined as

    .. math::
        \\text{mean error} =\\frac{1}{N}\\sum_{i=1}^{N}(x_i - y_i)
        \\text{where } x = \\text{the forecast, and } y = \\text{the observation}


    See "Mean error" section at https://www.cawcr.gov.au/projects/verification/ for more information

    Args:
        fcst: Forecast or predicted variables.
        obs: Observed variables.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the mean error. All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating the mean error. All other dimensions will be reduced. As a
            special case, 'all' will allow all dimensions to be preserved. In
            this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        An xarray object with the mean error of a forecast.

    """
    return additive_bias(fcst, obs, reduce_dims=reduce_dims, preserve_dims=preserve_dims, weights=weights)


def additive_bias(
    fcst: XarrayLike,
    obs: XarrayLike,
    *,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[XarrayLike] = None,
) -> XarrayLike:
    """
    Calculates the additive bias which is also sometimes called the mean error.

    It is defined as

    .. math::
        \\text{Additive bias} =\\frac{1}{N}\\sum_{i=1}^{N}(x_i - y_i)
        \\text{where } x = \\text{the forecast, and } y = \\text{the observation}


    See "Mean error" section at https://www.cawcr.gov.au/projects/verification/ for more information

    Args:
        fcst: Forecast or predicted variables.
        obs: Observed variables.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the additive bias. All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating the additive bias. All other dimensions will be reduced. As a
            special case, 'all' will allow all dimensions to be preserved. In
            this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        An xarray object with the additive bias of a forecast.

    """
    # Note - mean error call this function
    error = fcst - obs
    score = scores.functions.apply_weights(error, weights=weights)
    reduce_dims = scores.utils.gather_dimensions(
        fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
    )
    score = score.mean(dim=reduce_dims)
    return score  # type: ignore


def multiplicative_bias(
    fcst: XarrayLike,
    obs: XarrayLike,
    *,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[XarrayLike] = None,
) -> XarrayLike:
    """
    Calculates the multiplicative bias.

    Most suited for forecasts that have a lower bound at 0 such as wind speed. Will return
    a np.inf where the mean of `obs` across the dims to be reduced is 0.
    It is defined as

    .. math::
        \\text{{Multiplicative bias}} = \\frac{\\frac{1}{N}\\sum_{i=1}^{N}x_i}{\\frac{1}{N}\\sum_{i=1}^{N}y_i}
        \\text{where } x = \\text{the forecast, and } y = \\text{the observation}

    See "(Multiplicative) bias" section at https://www.cawcr.gov.au/projects/verification/ for more information

    Args:
        fcst: Forecast or predicted variables.
        obs: Observed variables.
        reduce_dims: Optionally specify which dimensions to reduce when
            calculating the multiplicative bias. All other dimensions will be preserved.
        preserve_dims: Optionally specify which dimensions to preserve when
            calculating the multiplicative bias. All other dimensions will be reduced. As a
            special case, 'all' will allow all dimensions to be preserved. In
            this case, the result will be in the same shape/dimensionality
            as the forecast, and the errors will be the error at each
            point (i.e. single-value comparison against observed), and the
            forecast and observed dimensions must match precisely.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)

    Returns:
        An xarray object with the multiplicative bias of a forecast.

    """
    reduce_dims = scores.utils.gather_dimensions(
        fcst.dims, obs.dims, reduce_dims=reduce_dims, preserve_dims=preserve_dims
    )
    fcst = scores.functions.apply_weights(fcst, weights=weights)
    obs = scores.functions.apply_weights(obs, weights=weights)

    # Need to broadcast and match NaNs so that the fcst mean and obs mean are for the
    # same points
    fcst, obs = broadcast_and_match_nan(fcst, obs)
    multi_bias = fcst.mean(dim=reduce_dims) / obs.mean(dim=reduce_dims)
    return multi_bias



"""Tools for matching data for verification"""

import xarray as xr

from scores.typing import XarrayLike


def broadcast_and_match_nan(*args: XarrayLike) -> tuple[XarrayLike, ...]:
    """
    Input xarray data objects are 'matched' - they are broadcast against each
    other (forced to have the same dimensions), and the position of nans are
    forced onto all DataArrays. This matching process is applied across all
    supplied DataArrays, as well as all DataArrays inside supplied Datasets.

    Args:
        *args: any number of xarray data objects supplied as positional arguments. See
            examples below.

    Returns:
        A tuple of data objects of the same length as the number of data objects
        supplied as input. Each returned object is the 'matched' version of the
        input.

    Raises:
        ValueError: if any input args is not an xarray data
            object.

    Examples:

        >>> # Matching xarray data objects
        >>> da1_matched, ds_matched, da2_matched = scores.processing.broadcast_and_match_nan(da1, ds, da2)

        >>> # Matching a tuple of xarray data objects
        >>> input_tuple = (da1, ds, da2)
        >>> matched_tuple = broadcast_and_match_nan(*input_tuple)
        >>> da1_matched = matched_tuple[0]
        >>> ds_matched = matched_tuple[1]
        >>> da2_matched = matched_tuple[2]
    """

    # sanitise inputs
    for i, arg in enumerate(args):
        if not isinstance(arg, (xr.Dataset, xr.DataArray)):
            raise ValueError(
                f"Argument {i} is not an xarray data object. (counting from 0, i.e. "
                "argument 0 is the first argument)"
            )

    # internal function to update the mask
    def update_mask(mask, data_array):
        """
        Perform the boolean AND operation on a mask (DataArray) and
        data_array.notnull()
        """
        return mask & data_array.notnull()

    # initialise the mask
    mask = True
    # generate the mask
    for arg in args:
        # update the mask for a DataArray
        if isinstance(arg, xr.DataArray):
            mask = update_mask(mask, arg)
        # update the mask for Datasets
        else:
            for data_var in arg.data_vars:
                mask = update_mask(mask, arg[data_var])

    # return matched data objects
    return tuple(arg.where(mask) for arg in args)



"""
This module supports the implementation of the CRPS scoring function, drawing from additional functions.
The two primary methods, `crps_cdf` and `crps_for_ensemble` are imported into 
the probability module to be part of the probability API.
"""

from collections.abc import Iterable
from typing import Literal, Optional, Sequence

import numpy as np
import pandas as pd
import xarray as xr

import scores.utils
from scores.probability.checks import coords_increasing
from scores.processing.cdf import (
    add_thresholds,
    cdf_envelope,
    decreasing_cdfs,
    integrate_square_piecewise_linear,
    observed_cdf,
    propagate_nan,
)


# pylint: disable=too-many-arguments
# pylint: disable=too-many-branches
def check_crps_cdf_inputs(
    fcst,
    obs,
    threshold_dim,
    threshold_weight,
    fcst_fill_method,
    threshold_weight_fill_method,
    integration_method,
    dims,
):
    """Checks that `crps_cdf` inputs are valid."""
    if threshold_dim not in fcst.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `fcst`")

    if threshold_weight is not None and threshold_dim not in threshold_weight.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `threshold_weight`")

    if threshold_dim in obs.dims:
        raise ValueError(f"'{threshold_dim}' is a dimension of `obs`")

    if not set(obs.dims).issubset(fcst.dims):
        raise ValueError("Dimensions of `obs` must be a subset of dimensions of `fcst`")

    if threshold_weight is not None and not set(threshold_weight.dims).issubset(fcst.dims):
        raise ValueError("Dimensions of `threshold_weight` must be a subset of dimensions of `fcst`")

    if dims is not None and not set(dims).issubset(fcst.dims):
        raise ValueError("`dims` must be a subset of `fcst` dimensions")  # pragma: no cover

    if fcst_fill_method not in ["linear", "step", "forward", "backward"]:
        raise ValueError("`fcst_fill_method` must be 'linear', 'step', 'forward' or 'backward'")

    if threshold_weight is not None and threshold_weight_fill_method not in [
        "linear",
        "step",
        "forward",
        "backward",
    ]:
        msg = "`threshold_weight_fill_method` must be 'linear', 'step', 'forward' or "
        msg += "'backward' if `threshold_weight` is supplied"
        raise ValueError(msg)

    if integration_method not in ["exact", "trapz"]:
        raise ValueError("`integration_method` must be 'exact' or 'trapz'")

    if len(fcst[threshold_dim]) < 2:
        raise ValueError("`threshold_dim` in `fcst` must have at least 2 values to calculate CRPS")

    if not coords_increasing(fcst, threshold_dim):
        raise ValueError("`threshold_dim` coordinates in `fcst` must be increasing")

    if threshold_weight is not None and not coords_increasing(threshold_weight, threshold_dim):
        raise ValueError("`threshold_dim` coordinates in `threshold_weight` must be increasing")

    if threshold_weight is not None and (threshold_weight < 0).any():
        raise ValueError("`threshold_weight` has negative values")


def crps_cdf_reformat_inputs(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    threshold_dim: str,
    *,  # Force keywords arguments to be keyword-only
    threshold_weight: Optional[xr.DataArray] = None,
    additional_thresholds: Optional[Iterable[float]] = None,
    fcst_fill_method: Literal["linear", "step", "forward", "backward"] = "linear",
    threshold_weight_fill_method: Literal["linear", "step", "forward", "backward"] = "forward",
) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    """
    Takes `fcst`, `obs` and `threshold_weight` inputs from `crps_cdf` and reformats them
    for use in `crps_cdf`; that is, `cdf_fcst`, `cdf_obs` and `threshold_weight` are
    returned with matching dimensions and coordinates.

    `additional_thresholds` contains additional thresholds to use.
    Useful if wanting to increase density of thresholds for `crps_cdf_trapz`.

    NaNs are filled, unless there are too few non-NaNs for filling.
    """
    # will use all thresholds from fcst, obs and (if applicable) weight

    fcst_thresholds = fcst[threshold_dim].values
    obs_thresholds = pd.unique(obs.values.flatten())

    weight_thresholds = []  # type: ignore
    if threshold_weight is not None:
        weight_thresholds = threshold_weight[threshold_dim].values  # type: ignore

    if additional_thresholds is None:
        additional_thresholds = []

    thresholds = np.concatenate((weight_thresholds, fcst_thresholds, obs_thresholds, additional_thresholds))  # type: ignore
    thresholds = np.sort(pd.unique(thresholds))
    thresholds = thresholds[~np.isnan(thresholds)]

    # get obs in cdf form with correct thresholds
    obs_cdf = observed_cdf(
        obs,
        threshold_dim,
        threshold_values=thresholds,
        include_obs_in_thresholds=False,  # thresholds already has rounded obs values
        precision=0,
    )

    # get fcst with correct thresholds
    fcst = add_thresholds(fcst, threshold_dim, thresholds, fcst_fill_method)

    # get weight with correct thresholds
    if threshold_weight is None:
        # the weight is 1
        weight_cdf = xr.full_like(fcst, 1.0)
    else:
        weight_cdf = add_thresholds(threshold_weight, threshold_dim, thresholds, threshold_weight_fill_method)

    fcst_cdf, obs_cdf, weight_cdf = xr.broadcast(fcst, obs_cdf, weight_cdf)

    return fcst_cdf, obs_cdf, weight_cdf


# pylint: disable=too-many-locals
def crps_cdf(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    threshold_dim: str = "threshold",
    threshold_weight: Optional[xr.DataArray] = None,
    additional_thresholds: Optional[Iterable[float]] = None,
    propagate_nans: bool = True,
    fcst_fill_method: Literal["linear", "step", "forward", "backward"] = "linear",
    threshold_weight_fill_method: Literal["linear", "step", "forward", "backward"] = "forward",
    integration_method: Literal["exact", "trapz"] = "exact",
    reduce_dims: Optional[Iterable[str]] = None,
    preserve_dims: Optional[Iterable[str]] = None,
    weights=None,
    include_components=False,
):
    """Calculates the CRPS probabilistic metric given CDF input.

    Calculates the continuous ranked probability score (CRPS), or the mean CRPS over
    specified dimensions, given forecasts in the form of predictive cumulative
    distribution functions (CDFs). Can also calculate threshold-weighted versions of the
    CRPS by supplying a `threshold_weight`.

    Predictive CDFs here are described by an indexed set of values rather than by
    closed forumulae. As a result, the resolution or number of points at which the CDF
    is realised has an impact on the calculation of areas under and over the curve to
    obtain the CRPS result. The term 'threshold' is used to describe the dimension which
    is used as an index for predictive CDF values. Various techniques are used to
    interpolate CDF values between indexed thresholds.

    Given
        - a predictive CDF `fcst` indexed at thresholds by variable x,
        - an observation in CDF form `obs_cdf` (i.e., obs_cdf(x) = 0 if x < obs and 1 if x >= obs),
        - a `threshold_weight` array indexed by variable x,

    The threshold-weighted CRPS is given by:
        `CRPS = integral(threshold_weight(x) * (fcst(x) - obs_cdf(x))**2)`, over all thresholds x.
    The usual CRPS is the threshold-weighted CRPS with `threshold_weight(x) = 1` for all x.
    This can be decomposed into an over-forecast penalty
        `integral(threshold_weight(x) * (fcst(x) - obs_cdf(x))**2)`, over all thresholds x where x >= obs
    and an under-forecast penalty
        `integral(threshold_weight(x) * (fcst(x) - obs_cdf(x))**2)`, over all thresholds x where x <= obs.

    Note that the function `crps_cdf` is designed so that the `obs` argument contains
    actual observed values. `crps_cdf` will convert `obs` into CDF form in order to
    calculate the CRPS.

    To calculate CRPS, integration is applied over the set of thresholds x taken from
        - `fcst[threshold_dim].values`,
        - `obs.values`.
        - `threshold_weight[threshold_dim].values` if applicable.
        - `additional_thresholds` if applicable.
    With NaN values excluded. There are two methods of integration:
        - "exact" gives the exact integral under that assumption that that `fcst` is
          continuous and piecewise linear between its specified values, and that
          `threshold_weight` (if supplied) is piecewise constant and right-continuous
          between its specified values.
        - "trapz" simply uses a trapezoidal rule using the specified values, and so is
          an approximation of the CRPS. To get an accurate approximation, the density
          of threshold values can be increased by supplying `additional_thresholds`.

    Both methods of calculating CRPS may require adding additional values to the
    `threshold_dim` dimension in `fcst` and (if supplied) `threshold_weight`.
    `fcst_fill_method` and `weight_fill_method` specify how `fcst` and `weight` are to
    be filled at these additional points.

    The function `crps_cdf` calculates the CRPS using forecast CDF values 'as is'.
    No checks are performed to ensure that CDF values in `fcst` are nondecreasing along
    `threshold_dim`. Checks are conducted on `fcst` and `threshold_weight` (if applicable)
    to ensure that coordinates are increasing along `threshold_dim`.

    Args:
        fcst: array of forecast CDFs, with the threshold dimension given by `threshold_dim`.
        obs: array of observations, not in CDF form.
        threshold_dim: name of the dimension in `fcst` that indexes the thresholds.
        threshold_weight: weight to be applied along `threshold_dim` to calculat
            threshold-weighted CRPS. Must contain `threshold_dim` as a dimension, and may
            also include other dimensions from `fcst`. If `weight=None`, a weight of 1
            is applied everywhere, which gives the usual CRPS.
        additional_thresholds: additional thresholds values to add to `fcst` and (if
            applicable) `threshold_weight` prior to calculating CRPS.
        propagate_nans: If `True`, propagate NaN values along `threshold_dim` in `fcst`
            and `threshold_weight` prior to calculating CRPS. This will result in CRPS
            being NaN for these cases. If `False`, NaN values in `fcst` and `weight` will
            be replaced, wherever possible, with non-NaN values using the fill method
            specified by `fcst_fill_method` and `threshold_weight_fill_method`.
        fcst_fill_method: how to fill values in `fcst` when NaNs have been introduced
            (by including additional thresholds) or are specified to be removed (by
            setting `propagate_nans=False`). Select one of:
                - "linear": use linear interpolation, then replace any leading or
                  trailing NaNs using linear extrapolation. Afterwards, all values are
                  clipped to the closed interval [0, 1].
                - "step": apply forward filling, then replace any leading NaNs with 0.
                - "forward": first apply forward filling, then remove any leading NaNs by
                  back filling.
                - "backward": first apply back filling, then remove any trailing NaNs by
                  forward filling.
            In most cases, "linear" is likely the appropriate choice.
        threshold_weight_fill_method: how to fill values in `threshold_weight` when NaNs
            have been introduced (by including additional thresholds) or are specified
            to be removed (by setting `propagate_nans=False`). Select one of "linear",
            "step", "forward" or "backward". If the weight function is continuous,
            "linear" is probably the best choice. If it is an increasing step function,
            "forward" may be best.
        integration_method (str): one of "exact" or "trapz".
        preserve_dims (Tuple[str]): dimensions to preserve in the output. All other dimensions are collapsed
            by taking the mean.
        reduce_dims (Tuple[str]): dimensions to reduce in the output by taking the mean. All other dimensions are
            preserved.
        weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
            by population, custom)
        include_components (bool): if True, include the under and over forecast components of
            the score in the returned dataset.

    Returns:
        xr.Dataset: The following are the produced Dataset variables:
            - "total" the total CRPS.
            - "underforecast_penalty": the under-forecast penalty contribution of the CRPS.
            - "overforecast_penalty": the over-forecast penalty contribution of the CRPS.

    Raises:
        ValueError: if `threshold_dim` is not a dimension of `fcst`.
        ValueError: if `threshold_dim` is not a dimension of `threshold_weight` when
            `threshold_weight` is not `None`.
        ValueError: if `threshold_dim` is a dimension of `obs`.
        ValueError: if dimensions of `obs` are not also dimensions of `fcst`.
        ValueError: if dimensions of `threshold_weight` are not also dimensions of `fcst`
            when `threshold_weight` is not `None`.
        ValueError: if `dims` is not a subset of dimensions of `fcst`.
        ValueError: if `fcst_fill_method` is not one of 'linear', 'step', 'forward' or
            'backward'.
        ValueError: if `weight_fill_method` is not one of 'linear', 'step', 'forward' or
            'backward'.
        ValueError: if `fcst[threshold_dim]` has less than 2 values.
        ValueError: if coordinates in `fcst[threshold_dim]` are not increasing.
        ValueError: if `threshold_weight` is not `None` and coordinates in
            `threshold_weight[threshold_dim]` are not increasing.
        ValueError: if `threshold_weight` has negative values.

    See also:
        - `scores.probability.crps_cdf_brier_decomposition`
        - `scores.probability.crps_for_ensemble`

    References:
        - Matheson, J. E., and R. L. Winkler, 1976: Scoring rules for continuous probability distributions.
            Manage. Sci.,22, 1087–1095.
        - Gneiting, T., & Ranjan, R. (2011). Comparing Density Forecasts Using Threshold- and
            Quantile-Weighted Scoring Rules.
            Journal of Business & Economic Statistics, 29(3), 411–422. http://www.jstor.org/stable/23243806
    """

    dims = scores.utils.gather_dimensions(
        fcst.dims,
        obs.dims,
        reduce_dims=reduce_dims,
        preserve_dims=preserve_dims,
    )

    check_crps_cdf_inputs(
        fcst,
        obs,
        threshold_dim,
        threshold_weight,
        fcst_fill_method,
        threshold_weight_fill_method,
        integration_method,
        dims,
    )

    if propagate_nans:
        fcst = propagate_nan(fcst, threshold_dim)  # type: ignore
        if threshold_weight is not None:
            threshold_weight = propagate_nan(threshold_weight, threshold_dim)  # type: ignore

    fcst, obs, threshold_weight = crps_cdf_reformat_inputs(
        fcst,
        obs,
        threshold_dim,
        threshold_weight=threshold_weight,
        additional_thresholds=additional_thresholds,
        fcst_fill_method=fcst_fill_method,
        threshold_weight_fill_method=threshold_weight_fill_method,
    )

    if integration_method == "exact":
        result = crps_cdf_exact(
            fcst,
            obs,
            threshold_weight,
            threshold_dim,
            include_components=include_components,
        )

    if integration_method == "trapz":
        result = crps_cdf_trapz(
            fcst,
            obs,
            threshold_weight,
            threshold_dim,
            include_components=include_components,
        )

    weighted = scores.functions.apply_weights(result, weights=weights)

    dims.remove(threshold_dim)  # type: ignore

    result = weighted.mean(dim=dims)  # type: ignore

    return result


def crps_cdf_exact(
    cdf_fcst: xr.DataArray,
    cdf_obs: xr.DataArray,
    threshold_weight: xr.DataArray,
    threshold_dim: str,
    *,  # Force keywords arguments to be keyword-only
    include_components=False,
) -> xr.Dataset:
    """
    Calculates exact value of CRPS assuming that:
        - the forecast CDF is continuous piecewise linear, with join points given by
          values in `cdf_fcst`,
        - the observation CDF is right continuous with values in {0,1} given by `cdf_obs`,
        - the threshold weight function is right continuous with values in {0,1} given
          by `threshold_weight`.

    If these assumptions do not hold, it might be best to use `crps_approximate`, with a
    sufficiently high resolution along `threshold_dim`.

    This function assumes that `cdf_fcst`, `cdf_obs`, `threshold_weight` have same shape.
    Also assumes that values along the `threshold_dim` dimension are increasing.

    Returns:
        (xr.Dataset): Dataset with `threshold_dim` collapsed containing DataArrays with
        CRPS and its decomposition, labelled "total", "underforecast_penalty" and
        "overforecast_penalty". NaN is returned if there is a NaN in the corresponding
        `cdf_fcst`, `cdf_obs` or `threshold_weight`.
    """

    # identify where input arrays have no NaN, collapsing `threshold_dim`
    # Mypy doesn't realise the isnan and any come from xarray not numpy
    inputs_without_nan = (
        ~np.isnan(cdf_fcst).any(threshold_dim)  # type: ignore
        & ~np.isnan(cdf_obs).any(threshold_dim)  # type: ignore
        & ~np.isnan(threshold_weight).any(threshold_dim)  # type: ignore
    )

    # thresholds in the closure of the interval (i.e. including endpoints) where
    # weight is 1
    interval_where_weight_one = (threshold_weight == 1) | ((threshold_weight.shift(**{threshold_dim: 1})) == 1)  # type: ignore

    # thresholds in the closure of the interval where cdf_obs is 1
    interval_where_obs_one = cdf_obs == 1

    # thresholds in the closure of the interval where obs cdf is 0
    interval_where_obs_zero = (cdf_obs == 0) | ((cdf_obs.shift(**{threshold_dim: 1})) == 0)  # type: ignore

    # over-forecast penalty contribution to CRPS: integral(w(x) * (F(x) - 1)^2) where x >= obs
    over = (cdf_fcst - 1).where(interval_where_obs_one).where(interval_where_weight_one)
    over = integrate_square_piecewise_linear(over, threshold_dim)
    # If zero penalty, could be NaN. Replace with 0, then NaN using inputs_without_nan
    over = over.where(~np.isnan(over), 0).where(inputs_without_nan)

    # under-forecast penalty contribution to CRPS: integral(w(x) * F(x)^2) where x < obs
    under = (cdf_fcst).where(interval_where_obs_zero).where(interval_where_weight_one)
    under = integrate_square_piecewise_linear(under, threshold_dim)
    # If zero penalty, could be NaN. Replace with 0, then nan using inputs_without_nan
    under = under.where(~np.isnan(under), 0).where(inputs_without_nan)

    total = over + under
    result = total.to_dataset(name="total")

    if include_components:
        result = xr.merge(
            [
                total.rename("total"),
                under.rename("underforecast_penalty"),
                over.rename("overforecast_penalty"),
            ]
        )

    return result


def crps_cdf_brier_decomposition(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    threshold_dim: str = "threshold",
    additional_thresholds: Optional[Iterable[float]] = None,
    fcst_fill_method: Literal["linear", "step", "forward", "backward"] = "linear",
    reduce_dims: Optional[Iterable[str]] = None,
    preserve_dims: Optional[Iterable[str]] = None,
) -> xr.Dataset:
    """
    Given an array `fcst` of predictive CDF values indexed along `threshold_dim`, and
    an array `obs` of  observations, calculates the mean Brier score for each index along
    `threshold_dim`. Since the mean CRPS is the integral of the mean Brier score over
    all thresholds, this gives a threshold decomposition of the mean CRPS.

    If any there are any NaNs along the threshold dimension of `fcst`, then NaNs are
    propagated along this dimension prior to calculating the decomposition. If
    propagating NaNs is not desired, the user may first fill NaNs in `fcst` using
    `scores.probability.functions.fill_cdf`.

    Args:
        fcst (xr.DataArray): DataArray of CDF values with threshold dimension `threshold_dim`.
        obs (xr.DataArray): DataArray of observations, not in CDF form.
        threshold_dim (str): name of the threshold dimension in `fcst`.
        additional_thresholds (Optional[Iterable[float]]): additional thresholds
            at which to calculate the mean Brier score.
        fcst_fill_method (Literal["linear", "step", "forward", "backward"]): How to fill NaN
            values in `fcst` that arise from new user-supplied thresholds or thresholds derived
            from observations.
            - "linear": use linear interpolation, and if needed also extrapolate linearly.
              Clip to 0 and 1. Needs at least two non-NaN values for interpolation,
              so returns NaNs where this condition fails.
            - "step": use forward filling then set remaining leading NaNs to 0.
              Produces a step function CDF (i.e. piecewise constant).
            - "forward": use forward filling then fill any remaining leading NaNs with
              backward filling.
            - "backward": use backward filling then fill any remaining trailing NaNs with
              forward filling.
        dims: dimensions to preserve in the output. The dimension `threshold_dim` is always
            preserved, even if not specified here.

    Returns:
        An xarray Dataset with data_vars:
            - "total_penalty": the mean Brier score for each threshold.
            - "underforecast_penalty": the mean of the underforecast penalties for the Brier score. For a particular
                forecast case, this component equals 0 if the event didn't occur and the Brier score if it did.
            - "overforecast_penalty": the mean of the overforecast penalties for the Brier score. For a particular
                forecast case, this component equals 0 if the event did occur and the Brier score if it did not.

    Raises:
        ValueError: if `threshold_dim` is not a dimension of `fcst`.
        ValueError: if `threshold_dim` is a dimension of `obs`.
        ValueError: if dimensions of `obs` are not also among the dimensions of `fcst`.
        ValueError: if dimensions in `dims` is not among the dimensions of `fcst`.
        ValueError: if `fcst_fill_method` is not one of 'linear', 'step', 'forward' or 'backward'.
        ValueError: if coordinates in `fcst[threshold_dim]` are not increasing.
    """
    dims = scores.utils.gather_dimensions(
        fcst.dims,
        obs.dims,
        reduce_dims=reduce_dims,
        preserve_dims=preserve_dims,
    )

    check_crps_cdf_brier_inputs(fcst, obs, threshold_dim, fcst_fill_method, dims)

    fcst = propagate_nan(fcst, threshold_dim)  # type: ignore

    fcst, obs, _ = crps_cdf_reformat_inputs(
        fcst,
        obs,
        threshold_dim,
        threshold_weight=None,
        additional_thresholds=additional_thresholds,
        fcst_fill_method=fcst_fill_method,
        threshold_weight_fill_method="forward",
    )

    dims.remove(threshold_dim)  # type: ignore

    # brier score for each forecast case
    bscore = (fcst - obs) ** 2
    not_nan = ~np.isnan(bscore)

    # `obs` here is the empirical CDF of the observation
    # when `obs == 1` the observation was lower than the threshold considered
    # and hence the Brier score at this threshold penalises an over-forecast
    over = bscore.where(np.isclose(obs, 1), 0).where(not_nan).mean(dims)
    under = bscore.where(np.isclose(obs, 0), 0).where(not_nan).mean(dims)
    total = over + under

    result = xr.merge(
        [
            total.rename("total_penalty"),
            under.rename("underforecast_penalty"),
            over.rename("overforecast_penalty"),
        ]
    )

    return result


def check_crps_cdf_brier_inputs(fcst, obs, threshold_dim, fcst_fill_method, dims):
    """Checks inputs to 'crps_cdf_brier_decomposition'."""

    if threshold_dim not in fcst.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `fcst`")

    if threshold_dim in obs.dims:
        raise ValueError(f"'{threshold_dim}' is a dimension of `obs`")

    if not set(obs.dims).issubset(fcst.dims):
        raise ValueError("Dimensions of `obs` must be a subset of dimensions of `fcst`")

    if dims is not None and not set(dims).issubset(fcst.dims):
        raise ValueError("`dims` must be a subset of `fcst` dimensions")  # pragma: no cover

    if fcst_fill_method not in ["linear", "step", "forward", "backward"]:
        raise ValueError("`fcst_fill_method` must be 'linear', 'step', 'forward' or 'backward'")

    if not coords_increasing(fcst, threshold_dim):
        raise ValueError("`threshold_dim` coordinates in `fcst` must be increasing")


def adjust_fcst_for_crps(
    fcst: xr.DataArray,
    threshold_dim: str,
    obs: xr.DataArray,
    *,  # Force keywords arguments to be keyword-only
    decreasing_tolerance: float = 0,
    additional_thresholds: Optional[Iterable[float]] = None,
    fcst_fill_method: Literal["linear", "step", "forward", "backward"] = "linear",
    integration_method: Literal["exact", "trapz"] = "exact",
) -> xr.DataArray:
    """
    This function takes a forecast cumulative distribution functions (CDF) `fcst`.
    If `fcst` is not decreasing outside of specified tolerance, it returns `fcst`.
    Otherwise, the CDF envelope for `fcst` is computed, and the CDF from among
        - `fcst`,
        - the upper envelope, and
        - the lower envelope
    that has the higher (i.e. worse) CRPS is returned. In the event of a tie,
    preference is given in the order `fcst` then upper.
    See `scores.probability.functions.cdf_envelope` for details about the CDF envelope.

    The use case for this is when, either due to rounding or poor forecast process, the
    forecast CDF `fcst` fails to be nondecreasing. Rather than simply replacing `fcst`
    with NaN, `adjust_fcst_for_crps` returns a CDF for which CRPS can be calculated, but
    possibly with a predictive performance cost as measured by CRPS.

    Whether a CDF is decreasing outside specified tolerance is determined as follows.
    For each CDF in `fcst`, the sum of incremental decreases along the threshold dimension
    is calculated. For example, if the CDF values are
        [0, 0.4, 0.3, 0.9, 0.88, 1]
    then the sum of incremental decreases is -0.12. This CDF decreases outside specified
    tolerance if 0.12 > `decreasing_tolerance`.

    The adjusted array of forecast CDFs is determined as follows:
        - any NaN values in `fcst` are propagated along `threshold_dim` so that in each case
            the entire CDF is NaN;
        - any CDFs in `fcst` that are decreasing within specified tolerance are unchanged;
        - any CDFs in `fcst` that are decreasing outside specified tolerance are replaced with
            whichever of the upper or lower CDF envelope gives the highest CRPS, unless the original
            values give a higher CRPS in which case original values are kept.
    See `scores.probability.functions.cdf_envelope` for a description of the 'CDF envelope'.
    If propagating NaNs is not desired, the user may first fill NaNs in `fcst` using
    `scores.probability.functions.fill_cdf`.
    The CRPS for each forecast case is calculated using `crps`, with a weight of 1.

    Args:
        fcst: DataArray of CDF values with threshold dimension `threshold_dim`.
        threshold_dim: name of the threshold dimension in `fcst`.
        obs: DataArray of observations.
        decreasing_tolerance: nonnegative tolerance value.
        additional_thresholds: optional additional thresholds passed on to `crps` when
            calculating CRPS.
        fcst_fill_method: `fcst` fill method passed on to `crps` when calculating CRPS.
        integration_method: integration method passed on to `crps` when calculating CRPS.

    Returns:
        An xarray DataArray of possibly adjusted forecast CDFs, where adjustments are made
        to penalise CDFs that decrease outside tolerance.

    Raises:
        ValueError: If `threshold_dim` is not a dimension of `fcst`.
        ValueError: If `decreasing_tolerance` is negative.
    """
    if threshold_dim not in fcst.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `fcst`")

    if decreasing_tolerance < 0:
        raise ValueError("`decreasing_tolerance` must be nonnegative")

    fcst = propagate_nan(fcst, threshold_dim)  # type: ignore

    is_decreasing = decreasing_cdfs(fcst, threshold_dim, decreasing_tolerance)

    if not is_decreasing.any():
        return fcst

    fcst_env = cdf_envelope(fcst, threshold_dim)

    # dimensions to preserve when calculating CRPS
    crps_dims = [x for x in fcst_env.dims if x != threshold_dim]

    crps_fcst_env = crps_cdf(
        fcst_env,
        obs,
        threshold_dim=threshold_dim,
        additional_thresholds=additional_thresholds,
        fcst_fill_method=fcst_fill_method,
        integration_method=integration_method,
        preserve_dims=crps_dims,  # type: ignore
    )

    fcst_type_to_use = crps_fcst_env["total"].idxmax("cdf_type")

    fcst_to_return = fcst_env.where(fcst_env["cdf_type"] == fcst_type_to_use).max("cdf_type")

    fcst_to_return = fcst_to_return.where(is_decreasing).combine_first(fcst)

    return fcst_to_return


def crps_cdf_trapz(
    cdf_fcst: xr.DataArray,
    cdf_obs: xr.DataArray,
    threshold_weight: xr.DataArray,
    threshold_dim: str,
    *,  # Force keywords arguments to be keyword-only
    include_components=False,
) -> xr.Dataset:
    """
    Returns dataset with CRPS estimate and decomposition using a trapezoidal rule on
    points from the integrand
        threshold_weight(x) * (cdf_fcst(x) - cdf_obs(x)) ** 2

    The error in the estimate improves as the distance between adjacent thresholds in
    `threshold_dim` approaches zero.

    NaN is returned if there is a NaN in the corresponding `cdf_fcst`, `cdf_obs` or
    `threshold_weight`.
    """

    # identify where input arrays have no NaN, collapsing `threshold_dim`
    # mypy doesn't realise the isnan and any come from xarray not numpy
    inputs_without_nan = (
        ~np.isnan(cdf_fcst).any(dim=threshold_dim)  # type: ignore
        & ~np.isnan(cdf_obs).any(dim=threshold_dim)  # type: ignore
        & ~np.isnan(threshold_weight).any(dim=threshold_dim)  # type: ignore
    )

    # total error measured by CRPS
    total = (threshold_weight * (cdf_fcst - cdf_obs) ** 2).integrate(threshold_dim).where(inputs_without_nan)

    over = (cdf_obs * threshold_weight * (cdf_fcst - cdf_obs) ** 2).integrate(threshold_dim).where(inputs_without_nan)

    under = total - over

    result = total.to_dataset(name="total")

    if include_components:
        result = xr.merge(
            [
                total.rename("total"),
                under.rename("underforecast_penalty"),
                over.rename("overforecast_penalty"),
            ]
        )

    return result


def crps_step_threshold_weight(
    step_points: xr.DataArray,
    threshold_dim: str,
    *,  # Force keywords arguments to be keyword-only
    threshold_values: Optional[Iterable[float]] = None,
    steppoints_in_thresholds: bool = True,
    steppoint_precision: float = 0,
    weight_upper: bool = True,
) -> xr.DataArray:
    """Generates an array of weights based on DataArray step points.

    Creates an array of threshold weights, which can be used to calculate
    threshold-weighted CRPS, based on a step function. Applies a weight of 1 when
    step_point >= threshold, and a weight of 0 otherwise. Zeros and ones in the output
    weight function can be reversed by setting `weight_upper=False`.

    Args:
        step_points (xr.DataArray): points at which the weight function changes value from 0 to 1.
        threshold_dim (str): name of the threshold dimension in the returned array weight function.
        threshold_values (str): thresholds at which to calculate weights.
        steppoints_in_thresholds (bool): include `step_points` among the `threshold_dim` values.
        steppoint_precision (float): precision at which to round step_points prior to calculating the
            weight function. Select 0 for no rounding.
        weight_upper (bool): If true, returns a weight of 1 if step_point >= threshold, and a
            weight of 0 otherwise. If false, returns a weight of 0 if step_point >= threshold,
            and a weight of 1 otherwise.

    Returns:
        (xr.DataArray): Zeros and ones with the dimensions in `step_points`
        and an additional `threshold_dim` dimension.
    """

    weight = observed_cdf(
        step_points,
        threshold_dim,
        threshold_values=threshold_values,
        include_obs_in_thresholds=steppoints_in_thresholds,
        precision=steppoint_precision,
    )

    if not weight_upper:
        weight = 1 - weight

    return weight


def crps_for_ensemble(
    fcst: xr.DataArray,
    obs: xr.DataArray,
    ensemble_member_dim: str,
    *,  # Force keywords arguments to be keyword-only
    method: Literal["ecdf", "fair"] = "ecdf",
    reduce_dims: Optional[Sequence[str]] = None,
    preserve_dims: Optional[Sequence[str]] = None,
    weights: Optional[xr.DataArray] = None,
) -> xr.DataArray:
    """Calculates the CRPS probabilistic metric given ensemble input.

    Calculates the continuous ranked probability score (CRPS) given an ensemble of forecasts.
    An ensemble of forecasts can also be thought of as a random sample from the predictive
    distribution.

    Given an observation y, and ensemble member values {x_i} (for 1 <= i <= M), the CRPS is
    calculated by the formula
        CRPS({x_i}, y) = (1 / M) * sum(|x_i - y|) - (1 / 2 * K) * sum(|x_i - x_j|),
    where the first sum is iterated over 1 <= i <= M and the second sum is iterated over
    1 <= i <= M and 1 <= j <= M.

    The value of the constant K in this formula depends on the method.
        - If `method="ecdf"` then K = M ** 2. In this case the CRPS value returned is
            the exact CRPS value for the emprical cumulation distribution function
            constructed using the ensemble values.
        - If `method="fair"` then K = M * (M - 1). In this case the CRPS value returned
            is the approximated CRPS where the ensemble values can be interpreted as a
            random sample from the underlying predictive distribution. This interpretation
            stems from the formula CRPS(F, Y) = E|X - Y| - E|X - X'|/2, where X and X'
            are independent samples of the predictive distribution F, Y is the observation
            (possibly unknown) and E denotes the expectation. This choice of K gives an
            unbiased estimate for the second expectation.

    Args:
        fcst: Forecast data. Must have a dimension `ensemble_member_dim`.
        obs: Observation data.
        ensemble_member_dim: the dimension that specifies the ensemble member or the sample
            from the predictive distribution.
        method: Either "ecdf" or "fair".
        reduce_dims: Dimensions to reduce. Can be "all" to reduce all dimensions.
        preserve_dims: Dimensions to preserve. Can be "all" to preserve all dimensions.
        weights: Weights for calculating a weighted mean of individual scores.

    Returns:
        xarray object of (weighted mean) CRPS values.

    Raises:
        ValueError: when method is not one of "ecdf" or "fair".

    See also:
        `scores.probability.crps_cdf`

    References:
        - C. Ferro (2014), "Fair scores for ensemble forecasts", Q J R Meteorol Soc
            140(683):1917-1923.
        - T. Gneiting T and A. Raftery (2007), "Strictly proper scoring rules, prediction,
            and estimation", J Am Stat Assoc, 102(477):359-37.
        - M. Zamo and P. Naveau (2018), "Estimation of the Continuous Ranked Probability
            Score with Limited Information and Applications to Ensemble Weather Forecasts",
            Math Geosci 50:209-234, https://doi.org/10.1007/s11004-017-9709-7
    """
    if method not in ["ecdf", "fair"]:
        raise ValueError("`method` must be one of 'ecdf' or 'fair'")

    weights_dims = None
    if weights is not None:
        weights_dims = weights.dims

    dims_for_mean = scores.utils.gather_dimensions(
        fcst.dims,
        obs.dims,
        weights_dims=weights_dims,
        reduce_dims=reduce_dims,
        preserve_dims=preserve_dims,
        score_specific_fcst_dims=ensemble_member_dim,
    )

    ensemble_member_dim1 = scores.utils.tmp_coord_name(fcst)

    # calculate forecast spread contribution
    fcst_copy = fcst.rename({ensemble_member_dim: ensemble_member_dim1})  # type: ignore

    fcst_spread_term = abs(fcst - fcst_copy).sum(dim=[ensemble_member_dim, ensemble_member_dim1])  # type: ignore
    ens_count = fcst.count(ensemble_member_dim)
    if method == "ecdf":
        fcst_spread_term = fcst_spread_term / (2 * ens_count**2)
    if method == "fair":
        fcst_spread_term = fcst_spread_term / (2 * ens_count * (ens_count - 1))

    # calculate final CRPS for each forecast case
    fcst_obs_term = abs(fcst - obs).mean(dim=ensemble_member_dim)
    result = fcst_obs_term - fcst_spread_term

    # apply weights and take means across specified dims
    result = scores.functions.apply_weights(result, weights=weights).mean(dim=dims_for_mean)  # type: ignore

    return result  # type: ignore



"""
Import the functions from the implementations into the public API
"""
from scores.categorical.binary_impl import (
    probability_of_detection,
    probability_of_false_detection,
)
from scores.categorical.contingency_impl import (
    BasicContingencyManager,
    BinaryContingencyManager,
    EventOperator,
    ThresholdEventOperator,
)
from scores.categorical.multicategorical_impl import firm

__all__ = [
    "probability_of_detection",
    "probability_of_false_detection",
    "firm",
    "BasicContingencyManager",
    "BinaryContingencyManager",
    "ThresholdEventOperator",
    "EventOperator",
]



"""
Contains frequently-used functions of a general nature within scores
"""

import warnings
from collections.abc import Hashable, Iterable, Sequence
from typing import Optional, Union

import numpy as np
import pandas as pd
import xarray as xr

from scores.typing import FlexibleDimensionTypes, XarrayLike

WARN_ALL_DATA_CONFLICT_MSG = """
You are requesting to reduce or preserve every dimension by specifying the string 'all'.
In this case, 'all' is also a named dimension in your data, leading to an ambiguity.
In order to reduce or preserve the named data dimension, specify ['all'] as a list item
rather than relying on string interpretation. The program will continue to interpret the
string as an instruction to reduce or preserve every dimension.
"""


ERROR_SPECIFIED_NONPRESENT_PRESERVE_DIMENSION = """
You are requesting to preserve a dimension which does not appear in your data 
(fcst, obs or weights). It is ambiguous how to proceed therefore an exception has been
raised instead.
"""

ERROR_SPECIFIED_NONPRESENT_REDUCE_DIMENSION = """
You are requesting to reduce a dimension which does not appear in your data
(fcst, obs or weights). It is ambiguous how to proceed therefore an exception has been 
raised instead.
"""

ERROR_OVERSPECIFIED_PRESERVE_REDUCE = """
You have specified both preserve_dims and reduce_dims. This method doesn't know how
to properly interpret that, therefore an exception has been raised.
"""


class DimensionError(Exception):
    """
    Custom exception used when attempting to operate over xarray DataArray or
    Dataset objects that do not have compatible dimensions.
    """


def gather_dimensions(  # pylint: disable=too-many-branches
    fcst_dims: Iterable[Hashable],
    obs_dims: Iterable[Hashable],
    *,  # Force keywords arguments to be keyword-only
    weights_dims: Optional[Iterable[Hashable]] = None,
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    score_specific_fcst_dims: Optional[FlexibleDimensionTypes] = None,
) -> set[Hashable]:
    """
    Establish which dimensions to reduce when calculating errors but before taking means.

    Args:
        fcst: Forecast data
        obs: Observation data
        weights: Weights for calculating a weighted mean of scores
        reduce_dims: Dimensions to reduce. Can be "all" to reduce all dimensions.
        preserve_dims: Dimensions to preserve. Can be "all" to preserve all dimensions.
        score_specific_fcst_dims: Dimension(s) in `fcst` that are reduced to calculate individual scores.
            Must not appear as a dimension in `obs`, `weights`, `reduce_dims` or `preserve_dims`.
            e.g. the ensemble member dimension if calculating CRPS for ensembles, or the
            threshold dimension of calculating CRPS for CDFs.

    Returns:
        Set of dimensions over which to take the mean once the checks are passed.

    Raises:
        ValueError: when `preserve_dims and `reduce_dims` are both specified.
        ValueError: when `score_specific_fcst_dims` is not a subset of `fcst.dims`.
        ValueError: when `obs.dims`, `weights.dims`, `reduce_dims` or `preserve_dims`
            contains elements from `score_specific_fcst_dims`.
        ValueError: when `preserve_dims and `reduce_dims` contain elements not among dimensions
            of the data (`fcst`, `obs` or `weights`).

    """

    all_data_dims = set(fcst_dims).union(set(obs_dims))
    if weights_dims is not None:
        all_data_dims = all_data_dims.union(set(weights_dims))

    # Handle error conditions related to specified dimensions
    if preserve_dims is not None and reduce_dims is not None:
        raise ValueError(ERROR_OVERSPECIFIED_PRESERVE_REDUCE)

    # all_scoring_dims is the set of dims remaining after individual scores are computed.
    all_scoring_dims = all_data_dims.copy()

    # Handle error conditions related to specified dimensions
    specified_dims = preserve_dims or reduce_dims

    if specified_dims == "all":
        if "all" in all_data_dims:
            warnings.warn(WARN_ALL_DATA_CONFLICT_MSG)
    elif specified_dims is not None:
        if isinstance(specified_dims, str):
            specified_dims = [specified_dims]

    # Raise errors unless score_specific_fcst_dims are in fcst.dims only
    if score_specific_fcst_dims is not None:
        if isinstance(score_specific_fcst_dims, str):
            score_specific_fcst_dims = [score_specific_fcst_dims]
        if not set(score_specific_fcst_dims).issubset(set(fcst_dims)):
            raise ValueError("`score_specific_fcst_dims` must be a subset of `fcst` dimensions")
        if len(set(obs_dims).intersection(set(score_specific_fcst_dims))) > 0:
            raise ValueError("`obs.dims` must not contain any `score_specific_fcst_dims`")
        if weights_dims is not None:
            if len(set(weights_dims).intersection(set(score_specific_fcst_dims))) > 0:
                raise ValueError("`weights.dims` must not contain any `score_specific_fcst_dims`")
        if specified_dims is not None and specified_dims != "all":
            if len(set(specified_dims).intersection(set(score_specific_fcst_dims))) > 0:
                raise ValueError("`reduce_dims` and `preserve_dims` must not contain any `score_specific_fcst_dims`")

        # Finally, remove score_specific_fcst_dims from all_scoring_dims
        all_scoring_dims = all_data_dims.difference(set(score_specific_fcst_dims))

    if specified_dims is not None and specified_dims != "all":
        if not set(specified_dims).issubset(all_data_dims):
            if preserve_dims is not None:
                raise ValueError(ERROR_SPECIFIED_NONPRESENT_PRESERVE_DIMENSION)
            raise ValueError(ERROR_SPECIFIED_NONPRESENT_REDUCE_DIMENSION)

    # Turn preserve dims into reduce dims, if needed
    if preserve_dims is not None:
        if preserve_dims == "all":
            return set([])

        if isinstance(preserve_dims, str):
            preserve_dims = [preserve_dims]

        reduce_dims = set(all_scoring_dims).difference(preserve_dims)

    # Handle reduction of all dimensions by string "all""
    elif reduce_dims == "all":
        reduce_dims = set(all_scoring_dims)

    # Handle reduction of a single dim by string name
    elif isinstance(reduce_dims, str):
        reduce_dims = set([reduce_dims])

    # Handle when reduce_dims and preserve_dims are both None
    elif reduce_dims is None and preserve_dims is None:
        reduce_dims = set(all_scoring_dims)

    # Turn into a set if needed
    assert reduce_dims is not None  # nosec - this is just to modify type hinting
    reduce_dims = set(reduce_dims)

    # Reduce by list is the default so no handling needed
    return reduce_dims


def dims_complement(data, *, dims=None) -> list[str]:
    """Returns the complement of data.dims and dims

    Args:
        data: Input xarray object
        dims: an Iterable of strings corresponding to dimension names

    Returns:
        A sorted list of dimension names, the complement of data.dims and dims
    """

    if dims is None:
        dims = []

    # check that dims is in data.dims, and that dims is a of a valid form
    check_dims(data, dims, mode="superset")

    complement = set(data.dims) - set(dims)
    return sorted(list(complement))


def check_dims(xr_data: XarrayLike, expected_dims: Iterable[Hashable], *, mode: Optional[str] = None):
    """
    Checks the dimensions xr_data with expected_dims, according to `mode`.

    Args:
        xr_data: if a Dataset is supplied,
            all of its data variables (DataArray objects) are checked.
        expected_dims: an Iterable of dimension names.
        mode: one of 'equal' (default), 'subset' or 'superset'.
            If 'equal', checks that the data object has the same dimensions
            as `expected_dims`.
            If 'subset', checks that the dimensions of the data object is a
            subset of `expected_dims`.
            If 'superset', checks that the dimensions of the data object is a
            superset of `expected_dims`, (i.e. contains `expected_dims`).
            If 'proper subset', checks that the dimensions of the data object is a
            subset of `expected_dims`, (i.e. is a subset, but not equal to
            `expected_dims`).
            If 'proper superset', checks that the dimensions of the data object
            is a proper superset of `expected_dims`, (i.e. contains but is not
            equal to `expected_dims`).
            If 'disjoint', checks that the dimensions of the data object shares no
            elements with `expected_dims`.

    Raises:
        scores.utils.DimensionError: the dimensions of `xr_data` does
            not pass the check as specified by `mode`.
        TypeError: `xr_data` is not an xarray data object.
        ValueError: `expected_dims` contains duplicate values.
        ValueError: `expected_dims` cannot be coerced into a set.
        ValueError: `mode` is not one of 'equal', 'subset', 'superset',
            'proper subset', 'proper superset', or 'disjoint'
    """

    if isinstance(expected_dims, str):
        raise TypeError(f"Supplied dimensions '{expected_dims}' must be an iterable of strings, not a string itself.")

    try:
        dims_set = set(expected_dims)
    except Exception as exc:
        raise ValueError(
            f"Cannot convert supplied dims {expected_dims} into a set. Check debug log for more information."
        ) from exc

    if len(list(dims_set)) != len(list(expected_dims)):
        raise ValueError(f"Supplied dimensions {expected_dims} contains duplicate values.")

    if not hasattr(xr_data, "dims"):
        raise DimensionError("Supplied object has no dimensions")

    # internal functions to check a data array
    check_modes = {
        "equal": lambda da, dims_set: set(da.dims) == dims_set,
        "subset": lambda da, dims_set: set(da.dims) <= dims_set,
        "superset": lambda da, dims_set: set(da.dims) >= dims_set,
        "proper subset": lambda da, dims_set: set(da.dims) < dims_set,
        "proper superset": lambda da, dims_set: set(da.dims) > dims_set,
        "disjoint": lambda da, dims_set: len(set(da.dims) & dims_set) == 0,
    }

    if mode is None:
        mode = "equal"
    if mode not in check_modes:
        raise ValueError(f"No such mode {mode}, mode must be one of: {list(check_modes.keys())}")

    check_fn = check_modes[mode]

    # check the dims
    if not check_fn(xr_data, dims_set):
        raise DimensionError(
            f"Dimensions {list(xr_data.dims)} of data object are not {mode} to the "
            f"dimensions {sorted([str(d) for d in dims_set])}."
        )

    if isinstance(xr_data, xr.Dataset):
        # every data variable must pass the dims check too!
        for data_var in xr_data.data_vars:
            if not check_fn(xr_data[data_var], dims_set):
                raise DimensionError(
                    f"Dimensions {list(xr_data[data_var].dims)} of data variable "
                    f"'{data_var}' are not {mode} to the dimensions {sorted([str(d) for d in dims_set])}"
                )


def tmp_coord_name(xr_data: xr.DataArray, *, count=1) -> Union[str, list[str]]:
    """
    Generates temporary coordinate names that are not among the coordinate or dimension
    names of `xr_data`.

    Args:
        xr_data: Input xarray data array
        count: Number of unique names to generate

    Returns:
        If count = 1, a string which is the concatenation of 'new' with all coordinate and
        dimension names in the input array. (this is the default)
        If count > 1, a list of such strings, each unique from one another
    """
    all_names = ["new"] + list(xr_data.dims) + list(xr_data.coords)
    result = "".join(all_names)  # type: ignore

    if count == 1:
        return result

    results = [str(i) + result for i in range(count)]
    return results


def check_binary(data: XarrayLike, name: str):
    """
    Checks that data does not have any non-NaN values out of the set {0, 1}

    Args:
        data: The data to convert to check if only contains binary values
    Raises:
        ValueError: if there are values in `fcst` and `obs` that are not in the
            set {0, 1, np.nan} and `check_args` is true.
    """
    if isinstance(data, xr.DataArray):
        unique_values = pd.unique(data.values.flatten())
    else:
        unique_values = pd.unique(data.to_array().values.flatten())
    unique_values = unique_values[~np.isnan(unique_values)]
    binary_set = {0, 1}

    if not set(unique_values).issubset(binary_set):
        raise ValueError(f"`{name}` contains values that are not in the set {{0, 1, np.nan}}")



"""
Import the functions from the implementations into the public API
"""

from scores import continuous as __continuous
from scores.pandas.typing import PandasType


def mse(
    fcst: PandasType,
    obs: PandasType,
    *,  # Force keywords arguments to be keyword-only
    is_angular: bool = False,
) -> PandasType:
    """Calculates the mean squared error from forecast and observed data.

    A detailed explanation is on [Wikipedia](https://en.wikipedia.org/wiki/Mean_squared_error)


    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the base metric, `scores.continuous.mse`.

    Args:
        fcst: Forecast or predicted variables in pandas.
        obs: Observed variables in pandas.
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.

    Returns:
        An object containing
            a single floating point number representing the mean squared
            error for the supplied data. All dimensions will be reduced.

    """
    return __continuous.mse(fcst, obs, is_angular=is_angular)  # type: ignore  # mypy is wrong, I think


def rmse(
    fcst: PandasType,
    obs: PandasType,
    *,  # Force keywords arguments to be keyword-only
    is_angular: bool = False,
) -> PandasType:
    """Calculate the Root Mean Squared Error from xarray or pandas objects.

    A detailed explanation is on [Wikipedia](https://en.wikipedia.org/wiki/Root-mean-square_deviation)

    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the base metric, `scores.continuous.rmse`.

    Args:
        fcst: Forecast or predicted variables in pandas.
        obs: Observed variables in pandas.
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.

    Returns:
        An object containing
            a single floating point number representing the root mean squared
            error for the supplied data. All dimensions will be reduced.

    """
    return __continuous.rmse(fcst, obs, is_angular=is_angular)  # type: ignore  # mypy is wrong, I think


def mae(
    fcst: PandasType,
    obs: PandasType,
    *,  # Force keywords arguments to be keyword-only
    is_angular: bool = False,
) -> PandasType:
    """Calculates the mean absolute error from forecast and observed data.

    A detailed explanation is on [Wikipedia](https://en.wikipedia.org/wiki/Mean_absolute_error)


    Dimensional reduction is not supported for pandas and the user should
    convert their data to xarray to formulate the call to the base metric, `scores.continuous.mae`.

    Args:
        fcst: Forecast or predicted variables in pandas.
        obs: Observed variables in pandas.
        is_angular: specifies whether `fcst` and `obs` are angular
            data (e.g. wind direction). If True, a different function is used
            to calculate the difference between `fcst` and `obs`, which
            accounts for circularity. Angular `fcst` and `obs` data should be in
            degrees rather than radians.

    Returns:
        An object containing
            a single floating point number representing the mean absolute
            error for the supplied data. All dimensions will be reduced.

    """
    return __continuous.mae(fcst, obs, is_angular=is_angular)  # type: ignore  # mypy is wrong, I think



"""
Functions for calculating a modified Diebold-Mariano test statistic
"""
import warnings
from typing import Literal

import numpy as np
import scipy as sp
import xarray as xr
from scipy.optimize import least_squares

from scores.stats.statistical_tests.acovf import acovf
from scores.utils import dims_complement


def diebold_mariano(  # pylint: disable=R0914
    da_timeseries: xr.DataArray,
    ts_dim: str,
    h_coord: str,
    *,  # Force keywords arguments to be keyword-only
    method: Literal["HG", "HLN"] = "HG",
    confidence_level: float = 0.95,
    statistic_distribution: Literal["normal", "t"] = "normal",
) -> xr.Dataset:
    """
    Given an array of (multiple) timeseries, with each timeseries consisting of score
    differences for h-step ahead forecasts, calculates a modified Diebold-Mariano test
    statistic for each timeseries. Several other statistics are also returned such as
    the confidence that the population mean of score differences is greater than zero
    and confidence intervals for that mean.

    Two methods for calculating the test statistic have been implemented: the "HG"
    method Hering and Genton (2011) and the "HLN" method of Harvey, Leybourne and
    Newbold (1997). The default "HG" method has an advantage of only generating positive
    estimates for the spectral density contribution to the test statistic. For further
    details see `scores.stats.confidence_intervals.impl._dm_test_statistic`.

    Prior to any calculations, NaNs are removed from each timeseries. If there are NaNs
    in `da_timeseries` then a warning will occur. This is because NaNs may impact the
    autocovariance calculation.

    To determine the value of h for each timeseries of score differences of h-step ahead
    forecasts, one may ask 'How many observations of the phenomenon will be made between
    making the forecast and having the observation that will validate the forecast?'
    For example, suppose that the phenomenon is afternoon precipitation accumulation in
    New Zealand (00z to 06z each day). Then a Day+1 forecast issued at 03z on Day+0 will
    be a 2-ahead forecast, since Day+0 and Day+1 accumulations will be observed before
    the forecast can be validated. On the other hand, a Day+1 forecast issued at 09z on
    Day+0 will be a 1-step ahead forecast. The value of h for each timeseries in the
    array needs to be specified in one of the sets of coordinates.
    See the example below.

    Confidence intervals and "confidence_gt_0" statistics are calculated using the
    test statistic, which is assumed to have either the standard normal distribution
    or Student's t distribution with n - 1 degrees of freedom (where n is the length of
    the timeseries). The distribution used is specified by `statistic_distribution`. See
    Harvey, Leybourne and Newbold (1997) for why the t distribution may be preferred,
    especially for shorter timeseries.

    If `da_timeseries` is a chunked array, data will be brought into memory during
    this calculation due to the autocovariance implementation.

    Args:
        da_timeseries: a 2 dimensional array containing the timeseries.
        ts_dim: name of the dimension which identifies each timeseries in the array.
        h_coord: name of the coordinate specifying, for each timeseries, that the
            timeseries is an h-step ahead forecast. `h_coord` coordinates must be
            indexed by the dimension `ts_dim`.
        method: method for calculating the test statistic, one of "HG" or "HLN".
        confidence_level: the confidence level, between 0 and 1 exclusive, at which to
            calculate confidence intervals.
        statistic_distribution: the distribution of the test-statistic under the null
            hypothesis of equipredictive skill. Used to calculate the "confidence_gt_0"
            statistic and confidence intervals. One of "normal" or "t" (for Student's t
            distribution).

    Returns:
        Dataset, indexed by `ts_dim`, with six variables:
        - "mean": the mean value for each timeseries, ignoring NaNs
        - "dm_test_stat": the modified Diebold-Mariano test statistic for each
          timeseries
        - "timeseries_len": the length of each timeseries, with NaNs removed.
        - "confidence_gt_0": the confidence that the mean value of the population is
          greater than zero, based on the specified `statistic_distribution`.
          Precisely, it is the value of the cumululative distribution function
          evaluated at `dm_test_stat`.
        - "ci_upper": the upper end point of a confidence interval about the mean at
          specified `confidence_level`.
        - "ci_lower": the lower end point of a confidence interval about the mean at
          specified `confidence_level`.

    Raises:
        ValueError: if `method` is not one of "HG" or "HLN".
        ValueError: if `statistic_distribution` is not one of "normal" or "t".
        ValueError: if `0 < confidence_level < 1` fails.
        ValueError: if `len(da_timeseries.dims) != 2`.
        ValueError: if `ts_dim` is not a dimension of `da_timeseries`.
        ValueError: if `h_coord` is not a coordinate of `da_timeseries`.
        ValueError: if `ts_dim` is not the only dimension of
            `da_timeseries[h_coord]`.
        ValueError: if `h_coord` values aren't positive integers.
        ValueError: if `h_coord` values aren't less than the lengths of the
            timeseries after NaNs are removed.
        RuntimeWarnning: if there is a NaN in diffs.

    References:
        - Diebold and Mariano, 'Comparing predictive accuracy', Journal of Business and
          Economic Statistics 13 (1995), 253-265.
        - Hering and Genton, 'Comparing spatial predictions',
          Technometrics 53 no. 4 (2011), 414-425.
        - Harvey, Leybourne and Newbold, 'Testing the equality of prediction mean
          squared errors', International Journal of Forecasting 13 (1997), 281-291.

    Example:

        This array gives three timeseries of score differences.
        Coordinates in the "lead_day" dimension uniquely identify each timeseries.
        Here `ts_dim="lead_day"`.
        Coordinates in the "valid_date" dimension give the forecast validity timestamp
        of each item in the timeseries.
        The "h" coordinates specify that the timeseries are for 2, 3 and 4-step
        ahead forecasts respectively. Here `h_coord="h"`.

        >>> da_timeseries = xr.DataArray(
        ...     data=[[1, 2, 3.0, 4, np.nan], [2.0, 1, -3, -1, 0], [1.0, 1, 1, 1, 1]],
        ...     dims=["lead_day", "valid_date"],
        ...     coords={
        ...         "lead_day": [1, 2, 3],
        ...         "valid_date": ["2020-01-01", "2020-01-02", "2020-01-03", "2020-01-04", "2020-01-05"],
        ...         "h": ("lead_day", [2, 3, 4]),
        ...     },
        ... )

        >>> dm_test_stats(da_timeseries, "lead_day", "h")
    """
    if method not in ["HLN", "HG"]:
        raise ValueError("`method` must be one of 'HLN' or 'HG'.")

    if statistic_distribution not in ["normal", "t"]:
        raise ValueError("`statistic_distribution` must be one of 'normal' or 't'.")

    if not 0 < confidence_level < 1:
        raise ValueError("`confidence_level` must be strictly between 0 and 1.")

    if len(da_timeseries.dims) != 2:
        raise ValueError("`da_timeseries` must have exactly two dimensions.")

    if ts_dim not in da_timeseries.dims:
        raise ValueError(f"`ts_dim` '{ts_dim}' must be a dimension of `da_timeseries`.")

    if h_coord not in da_timeseries.coords:
        raise ValueError("`h_coord` must be among the coordinates of `da_timeseries`.")

    # the following will also catch NaNs in da_timeseries[h_coord]
    # It allows values like 7.0 to pass, which is OK for the application.
    if any(da_timeseries[h_coord].values % 1 != 0):
        raise ValueError("Every value in `da_timeseries[h_coord]` must be an integer.")

    if (da_timeseries[h_coord] <= 0).any():
        raise ValueError("Every value in `da_timeseries[h_coord]` must be positive.")

    other_dim = dims_complement(da_timeseries, dims=[ts_dim])[0]
    da_timeseries_len = da_timeseries.count(other_dim)

    if (da_timeseries_len <= da_timeseries[h_coord]).any():
        msg = "Each `h_coord` value must be less than the length of the corresponding timeseries"
        raise ValueError(msg + " after NaNs are removed")

    ts_dim_len = len(da_timeseries[ts_dim])
    test_stats = np.empty([ts_dim_len])
    ts_mean = da_timeseries.mean(other_dim).values

    for i in range(ts_dim_len):
        timeseries = da_timeseries.isel({ts_dim: i})
        h = int(timeseries[h_coord])
        test_stats[i] = _dm_test_statistic(timeseries.values, h, method=method)

    if statistic_distribution == "normal":
        pvals = sp.stats.norm.cdf(test_stats)
        ci_quantile = sp.stats.norm.ppf(1 - (1 - confidence_level) / 2)
    else:
        pvals = sp.stats.t.cdf(test_stats, da_timeseries_len.values - 1)
        ci_quantile = sp.stats.t.ppf(1 - (1 - confidence_level) / 2, da_timeseries_len.values - 1)

    result = xr.Dataset(
        data_vars={
            "mean": ([ts_dim], ts_mean),
            "dm_test_stat": ([ts_dim], test_stats),
            "timeseries_len": ([ts_dim], da_timeseries_len.values),
            "confidence_gt_0": ([ts_dim], pvals),
            "ci_upper": ([ts_dim], ts_mean * (1 + ci_quantile / test_stats)),
            "ci_lower": ([ts_dim], ts_mean * (1 - ci_quantile / test_stats)),
        },
        coords={ts_dim: da_timeseries[ts_dim].values},
    )

    return result


def _dm_test_statistic(diffs: np.ndarray, h: int, *, method: Literal["HG", "HLN"] = "HG") -> float:
    """
    Given a timeseries of score differences for h-step ahead forecasts, as a 1D numpy
    array, returns a modified Diebold-Mariano test statistic. NaNs are removed prior
    to computing the statistic.

    Two methods for computing the test statistic can be used: either the "HLN" method of
    Harvey, Leybourne and Newbold (1997), or "HG" method of Hering and Genton (2011).

    Both methods use a different technique for estimating the spectral density of
    `diffs` at frequency 0, compared with Diebold and Mariano (1995). The HLN method
    uses an improved and less biased estimate (see V_hat (see Equation (5) in Harvey)).
    However, this estimate can sometimes be nonpositive, in which case NaN is returned.

    The HG method estimates the spectral density component using an exponential model
    for the autocovariances of `diffs`, so that positivity is guaranteed.
    Hering and Genton (2011) fit model parameters using empirical autocovariances
    computed up to half of the maximum lag. In this implementation, empirical
    autocovariances are computed up to half of the maximum lag or a lag of `h`,
    whichever is larger. Model parameters are computed using
    `scipy.optimize.least_squares`. It is assumed that the two model parameters (sigma
    and theta, in the notation of Hering and Genton (2011)) are positive.

    In both methods, if the `diff` sequence consists only of 0 values then NaN is
    returned.

    Args:
        diffs: timeseries of score difference as a 1D numpy array, assumed not all NaN.
        h: integer indicating that forecasts are h-step ahead, assumed to be positive and
            less than the length of the timeseries with NaNs removed.
        method: the method for computing the test statistic, either "HG" or "HLN".

    Returns:
        Modified Diebold-Mariano test statistic for sequence of score differences, with
        NaNs removed.

    Raises:
        ValueError: if `method` is not one of "HLN" or "HG".
        ValueError: if `0 < h < len(diffs)` fails after NaNs removed.
        RuntimeWarnning: if there is a NaN in diffs.

    References:
        - Diebold and Mariano, 'Comparing predictive accuracy', Journal of Business and
          Economic Statistics 13 (1995), 253-265.
        - Hering and Genton, 'Comparing spatial predictions',
          Technometrics 53 no. 4 (2011), 414-425.
        - Harvey, Leybourne and Newbold, 'Testing the equality of prediction mean
          squared errors', International Journal of Forecasting 13 (1997), 281-291.
    """
    if method not in ["HLN", "HG"]:
        raise ValueError("`method` must be one of 'HLN' or 'HG'.")

    if np.isnan(np.sum(diffs)):
        warnings.warn(
            RuntimeWarning(
                "A least one NaN value was detected in `da_timeseries`. This may impact the "
                "calculation of autocovariances."
            )
        )

    diffs = diffs[~np.isnan(diffs)]

    if not 0 < h < len(diffs):
        raise ValueError("The condition `0 < h < len(diffs)`, after NaNs removed, failed.")

    nonzero_diffs = diffs[diffs != 0]

    if len(nonzero_diffs) == 0:
        test_stat = np.nan
    elif method == "HLN":
        test_stat = _hln_method_stat(diffs, h)
    else:  # method == 'HG'
        test_stat = _hg_method_stat(diffs, h)

    return test_stat  # type: ignore


def _hg_func(pars: list, lag: np.ndarray, acv: np.ndarray) -> np.ndarray:
    """
    Function whose values are to be minimised as part of the HG method for estimating
    the spectral density at 0.

    Args:
        pars: list of two model parameters (sigma and theta in the notation of Hering
            and Genton 2011)
        lag: 1D numpy array of the form [0, 1, 2, ..., n - 1], where n is the length
            of `acv`
        acv: 1D numpy array of empirical autocovariances with lags corresponding
            to `lag`

    Returns:
        Difference between modelled and empirical autocoveriances.

    References:
        Hering and Genton, 'Comparing spatial predictions',
        Technometrics 53 no. 4 (2011), 414-425.
    """
    return (pars[0] ** 2) * np.exp(-3 * lag / pars[1]) - acv  # type: ignore


def _hg_method_stat(diffs: np.ndarray, h: int) -> float:
    """
    Calculates the modified Diebold-Mariano test statistic using the "HG" method.
    Assumes that h < len(diffs).

    Args:
        diffs: a single (1D array) timeseries of score differences with NaNs removed.
        h: integer indicating that forecasts are h-step ahead, assumed to be positive
            and less than the length of the timeseries with NaNs removed.

    Returns:
        Diebold-Mariano test statistic using the HG method.
    """

    n = len(diffs)

    # use an exponential model for autocovariances of `diffs`
    max_lag = int(max(np.floor((n - 1) / 2), h))
    sample_autocvs = acovf(diffs)[0:max_lag]
    sample_lags = np.arange(max_lag)
    model_params = least_squares(_hg_func, [1, 1], args=(sample_lags, sample_autocvs), bounds=(0, np.inf)).x

    # use the model autocovariances to estimate spectral density at 0
    all_lags = np.arange(n)
    model_autocovs = (model_params[0] ** 2) * np.exp(-3 * all_lags / model_params[1])
    density_estimate = model_autocovs[0] + 2 * np.sum(model_autocovs[1:])
    test_stat = np.mean(diffs) / np.sqrt(density_estimate / n)

    return test_stat  # type: ignore


def _hln_method_stat(diffs: np.ndarray, h: int) -> float:
    """
    Given a timeseries of score differences for h-step ahead forecasts, as a 1D numpy
    array without NaNs, returns the modified Diebold-Mariano test statistic of
    Harvey et al (1997).

    If the value V_hat (see Equation (5) in Harvey) is nonpositive then NaN is returned.

    Args:
        diffs: timeseries of score difference as a 1D numpy array without any NaNs.
        h: integer indicating that forecasts are h-step ahead, assumed to be positive
            and less than the length of the timeseries with NaNs removed.

    Returns:
        Diebold-Mariano test statistic using the HLN method.
    """
    n = len(diffs)
    diffs_bar = np.mean(diffs)

    # Harvey (1997) Equation (3)
    test_stat = diffs_bar / _dm_v_hat(diffs, diffs_bar, n, h) ** 0.5

    # Harvey (1997) Equation (9)
    correction_factor = (n + 1 - 2 * h + h * (h - 1) / n) / n
    test_stat = (correction_factor**0.5) * test_stat

    return test_stat  # type: ignore


def _dm_gamma_hat_k(diffs: np.ndarray, diffs_bar: float, n: int, k: int) -> float:
    """
    Computes the quantity (n - k) * gamma_hat_star_k of Equation (5) in
    Harvey et al (1997).

    Args:
        diffs: a single timeseries of score differences with NaNs removed.
        diffs_bar: mean of diffs.
        n: length of diffs.
        k: integer between 1 and n-1 (inclusive), where n = len(diffs)

    Returns:
        The quantity (n - k) * gamma_hat_star_k.
    """
    prod = (diffs[k:n] - diffs_bar) * (diffs[0 : n - k] - diffs_bar)

    return np.sum(prod)  # type: ignore


def _dm_v_hat(diffs: np.ndarray, diffs_bar: float, n: int, h: int) -> float:
    """
    Computes the the quantity V_hat(d_bar) of Equation (5) in Harvey et al (1997).

    Args:
        diffs: a single timeseries of score differences with NaNs removed.
        diffs_bar: mean of diffs.
        n: length of diffs.
        h: integer between 1 and n - 1 (inclusive), where n = len(diffs)

    Returns:
        The quantity V_hat(d_bar). If the result is not positive, NaN is returned.
    """
    summands = np.empty(h - 1)
    for k in range(h - 1):
        summands[k] = _dm_gamma_hat_k(diffs, diffs_bar, n, k + 1)

    result = (_dm_gamma_hat_k(diffs, diffs_bar, n, 0) + 2 * np.sum(summands)) / n**2

    if result <= 0:
        result = np.nan  # type: ignore

    return result  # type: ignore



"""
Estimate autocovariances

Barebones reimplementation of the `acovf` from `statsmodels.api.tsa.acovf`, 
for use only with `scores.stats.test.diebold_mariano`

Package: https://www.statsmodels.org/devel/

Code reference: https://github.com/statsmodels/statsmodels/blob/main/statsmodels/tsa/stattools.py

Notes:
    All type checking and other features have been removed, as they aren't needed for the 
    `diebold_mariano` function.

Why:
    Reduce dependant packages

## Source License

Copyright (C) 2006, Jonathan E. Taylor
All rights reserved.

Copyright (c) 2006-2008 Scipy Developers.
All rights reserved.

Copyright (c) 2009-2018 statsmodels Developers.
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of statsmodels nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL STATSMODELS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""

import numpy as np

# pylint: skip-file

__all__ = ["acovf"]


def _next_regular(target):
    """
    Find the next regular number greater than or equal to target.
    Regular numbers are composites of the prime factors 2, 3, and 5.
    Also known as 5-smooth numbers or Hamming numbers, these are the optimal
    size for inputs to FFTPACK.

    Target must be a positive integer.
    """
    if target <= 6:
        return target

    # Quickly check if it's already a power of 2
    if not (target & (target - 1)):
        return target

    match = float("inf")  # Anything found will be smaller
    p5 = 1
    while p5 < target:
        p35 = p5
        while p35 < target:
            # Ceiling integer division, avoiding conversion to float
            # (quotient = ceil(target / p35))
            quotient = -(-target // p35)
            # Quickly find next power of 2 >= quotient
            p2 = 2 ** ((quotient - 1).bit_length())

            N = p2 * p35
            if N == target:
                return N
            elif N < match:
                match = N
            p35 *= 3
            if p35 == target:
                return p35
        if p35 < match:
            match = p35
        p5 *= 5
        if p5 == target:
            return p5
    if p5 < match:
        match = p5
    return match


def acovf(x):
    """
    Estimate autocovariances.

    Args:
        x (array_like):
            Time series data. Must be 1d.

    Returns:
        (np.ndarray):
            The estimated autocovariances.

    References:
        [1] Parzen, E., 1963. On spectral analysis with missing observations
            and amplitude modulation. Sankhya: The Indian Journal of
            Statistics, Series A, pp.383-392.
    """

    xo = x - x.mean()

    n = len(x)

    d = n * np.ones(2 * n - 1)

    nobs = len(xo)
    n = _next_regular(2 * nobs + 1)
    Frf = np.fft.fft(xo, n=n)
    acov = np.fft.ifft(Frf * np.conjugate(Frf))[:nobs] / d[nobs - 1 :]
    acov = acov.real

    return acov



"""
This module contains a variety of functions which modify data in various ways to process data structures
to support probabilistic verification.
"""

from collections.abc import Iterable
from typing import Literal, Optional

import numpy as np
import pandas as pd
import xarray as xr

from scores.probability.checks import (
    cdf_values_within_bounds,
    check_nan_decreasing_inputs,
)
from scores.typing import XarrayLike


def round_values(array: xr.DataArray, rounding_precision: float, *, final_round_decpl: int = 7) -> xr.DataArray:
    """Round data array to specified precision.

    Rounding is done differently to `xarray.DataArray.round` or `numpy.round` where
    the number of decimal places is specified in those cases. Instead, here the rounding
    precision is specified as a float. The value is rounded to the nearest value that is
    divisible by `rounding_precision`.

    For example, 3.73 rounded to precision 0.2 is 3.8, and 37.3 rounded to precision 20
    is 40.

    Assumes that rounding_precision >=0, with 0 indicating no rounding to be performed.
    If rounding_precision > 0, a final round to `final_round_decpl` decimal places is performed
    to remove artefacts of python rounding process.

    Args:
        array (xr.DataArray): array of data to be rounded
        rounding_precision (float): rounding precision
        final_round_decpl (int): final round to specified number of decimal
            places when `rounding_precision` > 0.

    Returns:
        xr.DataArray: DataArray with rounded values.

    Raises:
        ValueError: If `rounding_precision` < 0.
    """
    if rounding_precision < 0:
        raise ValueError(f"rounding_precision '{rounding_precision}' is negative")

    if rounding_precision > 0:
        array = (array / rounding_precision).round() * rounding_precision
        array = array.round(decimals=final_round_decpl)

    return array


def propagate_nan(cdf: XarrayLike, threshold_dim: str) -> XarrayLike:
    """Propagates the NaN values from a "cdf" variable along the `threshold_dim`.

    Args:
        cdf (xr.DataArray): CDF values, so that P(X <= threshold) = cdf_value for
            each threshold in the `threshold_dim` dimension.
        threshold_dim (str): name of the threshold dimension in `cdf`.

    Returns:
        xr.DataArray: `cdf` variable with NaNs propagated.

    Raises:
        ValueError: If `threshold_dim` is not a dimension of `cdf`.
    """
    if threshold_dim not in cdf.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `cdf`")

    where_nan = xr.DataArray(np.isnan(cdf)).any(dim=threshold_dim)
    result = cdf.where(~where_nan, np.nan)
    return result


def observed_cdf(
    obs: xr.DataArray,
    threshold_dim: str,
    *,  # Force keywords arguments to be keyword-only
    threshold_values: Optional[Iterable[float]] = None,
    include_obs_in_thresholds: bool = True,
    precision: float = 0,
) -> xr.DataArray:
    """Returns a data array of observations converted into CDF format.

    Such that:
        returned_value = 0 if threshold < observation
        returned_value = 1 if threshold >= observation

    Args:
        obs (xr.DataArray): observations
        threshold_dim (str): name of dimension in returned array that contains the threshold values.
        threshold_values (Optional[Iterable[float]]): values to include among thresholds.
        include_obs_in_thresholds (bool): if `True`, include (rounded) observed values among thresholds.
        precision (float): precision applied to observed values prior to constructing the CDF and
            thresholds. Select 0 for highest precision (i.e. no rounding).

    Returns:
        xr.DataArray: Observed CDFs and thresholds in the `threshold_dim` dimension.

    Raises:
        ValueError: if `precision < 0`.
        ValueError: if all observations are NaN and no non-NaN `threshold_values`
            are not supplied.
    """
    if precision < 0:
        raise ValueError("`precision` must be nonnegative.")

    threshold_values_as_array = np.array(threshold_values)

    if np.isnan(obs).all() and (threshold_values is None or np.isnan(threshold_values_as_array).all()):
        raise ValueError("must include non-NaN observations in thresholds or supply threshold values")

    if precision > 0:
        obs = round_values(obs, precision)

    thresholds = threshold_values_as_array if threshold_values is not None else []

    if include_obs_in_thresholds:
        thresholds = np.concatenate((obs.values.flatten(), thresholds))  # type: ignore

    # remove any NaN
    thresholds = [x for x in thresholds if not np.isnan(x)]  # type: ignore

    # pandas.unique retains the original ordering whereas set() may not
    # pandas.unique no longer accepts a simple array as input
    thresholds = pd.Series(thresholds)
    thresholds = np.sort(pd.unique(thresholds))  # type: ignore

    da_thresholds = xr.DataArray(
        thresholds,
        dims=[threshold_dim],
        coords={threshold_dim: thresholds},
    )

    dab_obs, dab_thresholds = xr.broadcast(obs, da_thresholds)

    cdf = dab_thresholds >= dab_obs

    # convert to 0 and 1
    cdf = cdf.astype(float)

    cdf = cdf.where(~np.isnan(dab_obs))

    return cdf


def integrate_square_piecewise_linear(function_values: xr.DataArray, threshold_dim: str) -> xr.DataArray:
    """Calculates integral values and collapses `threshold_dim`.

    Calculates integral(F(t)^2), where
        - If t in a threshold value in `threshold_dim` then F(t) is in `function_values`,
        - F is piecewise linear between each of the t values in `threshold_dim`.
    Returns value of the integral with `threshold_dim` collapsed and other dimensions preserved.
    Returns NaN if there are less than two non-NaN function_values.

    This function assumes that
        - `threshold_dim` is a dimension of `function_values`
        - coordinates of `threshold_dim` are increasing.

    Args:
        function_values (xr.DataArray): array of function values F(t).
        threshold_dim (xr.DataArray): dimension along which to integrate.

    Returns:
        xr.DataArray: Integral values and `threshold_dim` collapsed.
    """

    # notation: Since F is piecewise linear we have
    # F(t) = mt + b, whenever x[i-1] <= t <= x[i].

    # difference in x
    diff_xs = function_values[threshold_dim] - function_values[threshold_dim].shift(**{threshold_dim: 1})  # type: ignore

    # difference in function values
    diff_ys = function_values - function_values.shift(**{threshold_dim: 1})  # type: ignore

    # gradients m
    m_values = diff_ys / diff_xs

    # y intercepts b
    b_values = function_values.shift(**{threshold_dim: 1})  # type: ignore

    # integral for x[i-1] <= t <= x[i]
    piece_integral = (
        (m_values**2) * (diff_xs**3) / 3 + m_values * b_values * (diff_xs**2) + (b_values**2) * diff_xs
    )

    # Need at least one non-NaN piece_integral to return float.
    # Note: need at least two non-NaN function values to get a non-NaN piece_integral.
    return piece_integral.sum(threshold_dim, min_count=1)


def add_thresholds(
    cdf: xr.DataArray,
    threshold_dim: str,
    new_thresholds: Iterable[float],
    fill_method: Literal["linear", "step", "forward", "backward", "none"],
    *,  # Force keywords arguments to be keyword-only
    min_nonnan: int = 2,
) -> xr.DataArray:
    """Takes a CDF data array with dimension `threshold_dim` and adds values from `new_thresholds`.

    The CDF is then filled to replace any NaN values.
    The array `cdf` requires at least 2 non-NaN values along `threshold_dim`.

    Args:
        cdf (xr.DataArray): array of CDF values.
        threshold_dim (str): name of the threshold dimension in `cdf`.
        new_thresholds (Iterable[float]): new thresholds to add to `cdf`.
        fill_method (Literal["linear", "step", "forward", "backward", "none"]): one of "linear",
            "step", "forward" or "backward", as described in `fill_cdf`. If no filling, set to "none".
        min_nonnan (int): passed onto `fill_cdf` for performing filling.

    Returns:
        xr.DataArray: Additional thresholds, and values at those thresholds
        determined by the specified fill method.
    """

    thresholds = np.concatenate((cdf[threshold_dim].values, new_thresholds))  # type: ignore
    thresholds = np.sort(pd.unique(thresholds))
    thresholds = thresholds[~np.isnan(thresholds)]

    da_thresholds = xr.DataArray(data=thresholds, dims=[threshold_dim], coords={threshold_dim: thresholds})

    da_cdf = xr.broadcast(cdf, da_thresholds)[0]

    if fill_method != "none":
        da_cdf = fill_cdf(da_cdf, threshold_dim, fill_method, min_nonnan)

    return da_cdf


def fill_cdf(
    cdf: xr.DataArray,
    threshold_dim: str,
    method: Literal["linear", "step", "forward", "backward"],
    min_nonnan: int,
) -> xr.DataArray:
    """Fills NaNs in a CDF of a real-valued random variable along `threshold_dim` with appropriate values between 0 and 1.

    Args:
        cdf (xr.DataArray): CDF values, where P(Y <= threshold) = cdf_value for each threshold in `threshold_dim`.
        threshold_dim (str): the threshold dimension in the CDF, along which filling is performed.
        method (Literal["linear", "step", "forward", "backward"]): one of
            - "linear": use linear interpolation, and if needed also extrapolate linearly. Clip to 0 and 1.
              Needs at least two non-NaN values for interpolation, so returns NaNs where this condition fails.
            - "step": use forward filling then set remaining leading NaNs to 0.
              Produces a step function CDF (i.e. piecewise constant).
            - "forward": use forward filling then fill any remaining leading NaNs with backward filling.
            - "backward": use backward filling then fill any remaining trailing NaNs with forward filling.
        min_nonnan (int): the minimum number of non-NaN entries required along `threshold_dim` for filling to
            be performed. All CDF values are set to `np.nan` where this condition fails.
            `min_nonnan` must be at least 2 for the "linear" method, and at least 1 for the other methods.

    Returns:
        xr.DataArray: Containing the same values as `cdf` but with NaNs filled.

    Raises:
        ValueError: If `threshold_dim` is not a dimension of `cdf`.
        ValueError: If `min_nonnan` < 1 when `method="step"` or if `min_nonnan` < 2 when `method="linear"`.
        ValueError: If `method` is not "linear", "step", "forward" or "backward".
        ValueError: If any non-NaN value of `cdf` lies outside the unit interval [0,1].

    """

    if method not in ["linear", "step", "forward", "backward"]:
        raise ValueError("`method` must be 'linear', 'step', 'forward' or 'backward'")

    if not cdf_values_within_bounds(cdf):
        raise ValueError("Input CDF has some values less than 0 or greater than 1.")

    if threshold_dim not in cdf.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `cdf`")

    if min_nonnan < 1 and method != "linear":
        raise ValueError(f"`min_nonnan` must be at least 1 when `method='{method}'`")

    if min_nonnan < 2 and method == "linear":
        raise ValueError("`min_nonnan` must be at least 2 when `method='linear'`")

    # set cdf values to be NaN where min_nonnan requirement fails
    nonnan_count = cdf.count(threshold_dim)
    cdf = cdf.where(nonnan_count >= min_nonnan)

    # NaN filling
    if method == "linear":
        cdf = cdf.interpolate_na(threshold_dim, method="linear", fill_value="extrapolate").clip(min=0, max=1)

    if method == "step":
        cdf = cdf.ffill(threshold_dim).fillna(0)
        # NaN cdfs will now be all zero, so bring back Nans
        cdf = cdf.where(nonnan_count >= min_nonnan)

    if method == "forward":
        cdf = cdf.ffill(threshold_dim).bfill(threshold_dim)

    if method == "backward":
        cdf = cdf.bfill(threshold_dim).ffill(threshold_dim)

    return cdf


def decreasing_cdfs(cdf: xr.DataArray, threshold_dim: str, tolerance: float) -> xr.DataArray:
    """A CDF of a real-valued random variable should be nondecreasing along threshold_dim.

    This is sometimes violated due to rounding issues or bad forecast process.
    `decreasing_cdfs` checks CDF values decrease beyond specified tolerance; that is,
    whenever the sum of the incremental decreases exceeds tolerarance.
    For example, if the CDF values are
        [0, 0.4, 0.3, 0.9, 0.88, 1]
    then the sum of incremental decreases is -0.12. Given a specified positive `tolerance`,
    the CDF values decrease beyond tolerance if the sum of incremental decreases < -`tolerance`.

    Intended use is for CDFs with increasing coordinates along `threshold_dim` dimension, and where
    either each CDF is always NaN or always non-NaN.

    Args:
        cdf (xr.DataArray): data array of CDF values
        threshold_dim (str): threshold dimension, such that P(Y < threshold) = cdf_value.
        tolerance (float): nonnegative tolerance value.

    Returns:
        xr.DataArray: Containing `threshold_dim` collapsed and values True if and only if
        the CDF is decreasing outside tolerance. If the CDF consists only of NaNs then
        the value is False.

    Raises:
        ValueError: If `threshold_dim` is not a dimension of `cdf`.
        ValueError: If `tolerance` is negative.
        ValueError: If coordinates are not increasing along `threshold_dim`.
        ValueError: If some, but not all, CDF values in `cdf` along `threshold_dim` are NaN.
    """
    check_nan_decreasing_inputs(cdf, threshold_dim, tolerance)

    # difference between consecutive terms along threshold_dim
    diff = cdf - cdf.shift(**{threshold_dim: 1})  # type: ignore

    result = diff.clip(max=0).sum(dim=threshold_dim) < -tolerance

    return result


def cdf_envelope(
    cdf: xr.DataArray,
    threshold_dim: str,
) -> xr.DataArray:
    """Forecast cumulative distribution functions (CDFs) for real-valued random variables.

    CDFs that are reconstructed from known points on the distribution should be nondecreasing
    with respect to the threshold dimension. However, sometimes this may fail due to rounding
    or poor forecast process. This function returns the "envelope" of the original CDF, which
    consists of two bounding CDFs, both of which are nondecreasing.

    The following example shows values from an original CDF that has a decreasing subsequence
    (and so is not a true CDF). The resulting "upper" and "lower" CDFs minimally adjust
    "original" so that "lower" <= "original" <= "upper".
        "original": [0, .5, .2, .8, 1]
        "upper": [0, .5, .5, .8, 1]
        "lower": [0, .2, .2, .8, 1]

    This function does not perform checks that `0 <= cdf <= 1`.

    Args:
        cdf (xr.DataArray): forecast CDF with thresholds in the thresholds_dim.
        threshold_dim (str): dimension in fcst_cdf that contains the threshold ordinates.

    Returns:
        An xarray DataArray consisting of three CDF arrays indexed along the `"cdf_type"` dimension
        with the following indices:
            - "original": same data as `cdf`.
            - "upper": minimally adjusted "original" CDF that is nondecreasing and
              satisfies "upper" >= "original".
            - "lower": minimally adjusted "original" CDF that is nondecreasing and
              satisfies "lower" <= "original".
        NaN values in `cdf` are maintained in "original", "upper" and "lower".

    Raises:
        ValueError: If `threshold_dim` is not a dimension of `cdf`.
    """
    if threshold_dim not in cdf.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `cdf`")

    # logic below assumes that cdf[threshold_dim].values are ascending
    cdf = cdf.sortby(threshold_dim)

    result = xr.full_like(cdf.expand_dims({"cdf_type": ["original", "upper", "lower"]}), np.nan)

    dim_idx = cdf.dims.index(threshold_dim)

    # use fmax so as not to propogate nans
    cdf_upper = np.fmax.accumulate(cdf.values, axis=dim_idx)
    cdf_lower = np.flip(
        1 - np.fmax.accumulate(1 - np.flip(cdf.values, axis=dim_idx), axis=dim_idx),
        axis=dim_idx,
    )

    result.loc["original"] = cdf.copy()
    result.loc["upper"] = np.where(~np.isnan(cdf), cdf_upper, np.nan)
    result.loc["lower"] = np.where(~np.isnan(cdf), cdf_lower, np.nan)

    return result



"""
This module contains methods related to the Brier score
"""

from typing import Optional

import xarray as xr

from scores.continuous import mse
from scores.typing import FlexibleDimensionTypes, XarrayLike
from scores.utils import check_binary


def brier_score(
    fcst: XarrayLike,
    obs: XarrayLike,
    *,  # Force keywords arguments to be keyword-only
    reduce_dims: Optional[FlexibleDimensionTypes] = None,
    preserve_dims: Optional[FlexibleDimensionTypes] = None,
    weights: Optional[xr.DataArray] = None,
    check_args: bool = True,
):
    """
    Calculates the Brier score for forecast and observed data. For an explanation of the
    Brier score, see [Wikipedia](https://en.wikipedia.org/wiki/Brier_score).

    If you want to speed up performance with Dask and are confident of your input data,
    or if you want observations to take values between 0 and 1, set `check_inputs` to `False`.

    Args:
            fcst: Forecast or predicted variables in xarray.
            obs: Observed variables in xarray.
            reduce_dims: Optionally specify which dimensions to reduce when
                calculating the Brier score. All other dimensions will be preserved.
            preserve_dims: Optionally specify which dimensions to preserve when
                calculating the Brier score. All other dimensions will be reduced. As a
                special case, 'all' will allow all dimensions to be preserved. In
                this case, the result will be in the same shape/dimensionality
                as the forecast, and the errors will be the absolute error at each
                point (i.e. single-value comparison against observed), and the
                forecast and observed dimensions must match precisely.
            weights: Optionally provide an array for weighted averaging (e.g. by area, by latitude,
                by population, custom)
    Raises:
        ValueError: if `fcst` contains non-nan values outside of the range [0, 1]
        ValueError: if `obs` contains non-nan values not in the set {0, 1}
    """
    if check_args:
        error_msg = ValueError("`fcst` contains values outside of the range [0, 1]")
        if isinstance(fcst, xr.Dataset):
            # The .values is required as item() is not yet a valid method on Dask
            if fcst.to_array().max().values.item() > 1 or fcst.to_array().min().values.item() < 0:
                raise error_msg
        else:
            if fcst.max().values.item() > 1 or fcst.min().values.item() < 0:
                raise error_msg
        check_binary(obs, "obs")

    return mse(fcst, obs, reduce_dims=reduce_dims, preserve_dims=preserve_dims, weights=weights)



"""
Import the functions from the implementations into the public API
"""

from scores.continuous.murphy_impl import murphy_score, murphy_thetas
from scores.probability.brier_impl import brier_score
from scores.probability.crps_impl import (
    adjust_fcst_for_crps,
    crps_cdf,
    crps_cdf_brier_decomposition,
    crps_for_ensemble,
)
from scores.probability.roc_impl import roc_curve_data
from scores.processing.isoreg_impl import isotonic_fit

__all__ = [
    "murphy_score",
    "murphy_thetas",
    "brier_score",
    "adjust_fcst_for_crps",
    "crps_cdf",
    "crps_cdf_brier_decomposition",
    "crps_for_ensemble",
    "roc_curve_data",
    "isotonic_fit",
]



"""
This module contains methods which make assertions at runtime about the state of various data
structures and values
"""

import numpy as np
import xarray as xr


def coords_increasing(da: xr.DataArray, dim: str):
    """Checks if coordinates in a given DataArray are increasing.

    Note: No in-built raise if `dim` is not a dimension of `da`.

    Args:
        da (xr.DataArray): Input data
        dim (str): Dimension to check if increasing
    Returns:
        (bool):  Returns True if coordinates along `dim` dimension of
        `da` are increasing, False otherwise.
    """
    result = (da[dim].diff(dim) > 0).all()
    return result


def cdf_values_within_bounds(cdf: xr.DataArray) -> bool:
    """Checks that 0 <= cdf <= 1. Ignores NaNs.

    Args:
        cdf (xr.DataArray): array of CDF values

    Returns:
        (bool): `True` if `cdf` values are all between 0 and 1 whenever values are not NaN,
            or if all values are NaN; and `False` otherwise.
    """
    flag = cdf.count() == 0 or ((cdf.min() >= 0) & (cdf.max() <= 1))
    return flag  # type: ignore  # mypy thinks flag could be a DataArray


def check_nan_decreasing_inputs(cdf, threshold_dim, tolerance):
    """Checks inputs to `nan_decreasing_cdfs` and `_decreasing_cdfs`."""

    if threshold_dim not in cdf.dims:
        raise ValueError(f"'{threshold_dim}' is not a dimension of `cdf`")

    if tolerance < 0:
        raise ValueError("`tolerance` must be nonnegative.")

    if not coords_increasing(cdf, threshold_dim):
        raise ValueError(f"Coordinates along '{threshold_dim}' dimension should be increasing.")

    all_nan_or_no_nan = np.isnan(cdf).all(threshold_dim) | (~np.isnan(cdf)).all(threshold_dim)
    if not all_nan_or_no_nan.all():
        raise ValueError("CDFs should have no NaNs or be all NaN along `threshold_dim`")



"""
Import the functions from the implementations into the public API
"""

from scores.processing.discretise import (
    binary_discretise,
    binary_discretise_proportion,
    comparative_discretise,
    proportion_exceeding,
)
from scores.processing.isoreg_impl import isotonic_fit
from scores.processing.matching import broadcast_and_match_nan

__all__ = [
    "comparative_discretise",
    "binary_discretise",
    "proportion_exceeding",
    "broadcast_and_match_nan",
    "binary_discretise_proportion",
    "isotonic_fit",
]
